"""Selection list."""

from __future__ import annotations

from natsort import index_natsorted, order_by_index
from qtpy.QtCore import QSize, Qt, Signal
from qtpy.QtGui import QAbstractTextDocumentLayout, QPalette, QTextDocument
from qtpy.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QFormLayout,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QStyle,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QWidget,
)

import qtextra.helpers as hp
from qtextra.widgets.qt_toolbar_mini import QtMiniToolbar


class HTMLDelegate(QStyledItemDelegate):
    """Rich text delegate."""

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        self.doc = QTextDocument(self)

    def paint(self, painter, option, index):
        painter.save()
        options = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)
        self.doc.setHtml(options.text)
        options.text = ""
        style = QApplication.style() if options.widget is None else options.widget.style()
        style.drawControl(QStyle.CE_ItemViewItem, options, painter)

        ctx = QAbstractTextDocumentLayout.PaintContext()
        if option.state & QStyle.State_Selected:
            ctx.palette.setColor(QPalette.Text, option.palette.color(QPalette.Active, QPalette.HighlightedText))
        else:
            ctx.palette.setColor(QPalette.Text, option.palette.color(QPalette.Active, QPalette.Text))
        textRect = style.subElementRect(QStyle.SE_ItemViewItemText, options, None)
        if index.column() != 0:
            textRect.adjust(5, 0, 0, 0)
        constant = 4
        margin = (option.rect.height() - options.fontMetrics.height()) // 2
        margin = margin - constant
        textRect.setTop(textRect.top() + margin)

        painter.translate(textRect.topLeft())
        painter.setClipRect(textRect.translated(-textRect.topLeft()))
        self.doc.documentLayout().draw(painter, ctx)
        painter.restore()

    def sizeHint(self, option, index):
        """Size hint."""
        return QSize(self.doc.idealWidth(), self.doc.size().height())


class QtSelectionList(QWidget):
    """Widget which allows to select items from a list.

    It also provides an easy to use interface to add items to the list, select all, deselect all and invert selection.
    """

    evt_selection_changed = Signal()

    _layout: QFormLayout
    list_widget: QListWidget
    toolbar: QtMiniToolbar
    filter_by: QLineEdit

    def __init__(
        self,
        parent: QWidget | None = None,
        allow_toolbar: bool = True,
        allow_sort: bool = True,
        allow_filter: bool = True,
        allow_visible_toggle: bool = True,
        enable_single_click: bool = False,
        double_click_to_select: bool = True,
    ):
        super().__init__(parent)
        self.allow_toolbar = allow_toolbar
        self.allow_sort = allow_sort
        self.allow_filter = allow_filter
        self.allow_visible_toggle = allow_visible_toggle

        # actions
        self.enable_single_click = enable_single_click
        self.double_click_to_select = double_click_to_select
        self.init_ui()

    # noinspection PyAttributeOutsideInit
    def init_ui(self) -> None:
        """Initialize the user interface."""
        self._layout = hp.make_form_layout(parent=self)

        self.filter_by = hp.make_line_edit(
            self, placeholder="Type in text to filter...", func_changed=self.on_filter, func_clear=self.on_filter
        )
        self.filter_by.setMinimumWidth(200)
        hp.set_expanding_sizer_policy(self.filter_by, horz=True)
        if not self.allow_filter:
            self.filter_by.hide()

        self.info_label = hp.make_label(self, "")

        self.toolbar = QtMiniToolbar(self, add_spacer=False, spacing=2, icon_size="average")
        self.toolbar.add_qta_tool("toggle_on", func=self.on_select_all, average=True, tooltip="Select all items")
        self.toolbar.add_qta_tool("toggle_off", func=self.on_deselect_all, average=True, tooltip="Deselect all items")
        self.toolbar.add_qta_tool(
            "invert_selection",
            func=self.on_invert_selection,
            average=True,
            tooltip="Invert selection",
            hide=not self.allow_visible_toggle,
        )
        self.toolbar.add_qta_tool("visible", func=self.on_show_all, average=True, tooltip="Show all items")
        self.toolbar.add_qta_tool(
            "visible_off",
            func=self.on_hide_deselected,
            average=True,
            tooltip="Hide deselected items",
            hide=not self.allow_visible_toggle,
        )
        self.toolbar.add_widget(self.filter_by, stretch=True)
        self.toolbar.add_widget(self.info_label)
        self.toolbar.max_size = self.filter_by.minimumHeight() + 2

        if not self.allow_filter:
            self.toolbar.add_spacer()
        if not self.allow_toolbar:
            self.toolbar.hide()
        self._layout.addRow(self.toolbar)

        self.list_widget = QListWidget(self)
        self.list_widget.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
        self.list_widget.setWordWrap(True)
        self.list_widget.itemChanged.connect(self.on_selection_changed)
        if self.double_click_to_select:
            self.list_widget.itemDoubleClicked.connect(
                lambda item: item.setCheckState(
                    Qt.CheckState.Checked if item.checkState() == Qt.CheckState.Unchecked else Qt.CheckState.Unchecked
                )
            )
        if self.enable_single_click:
            self.list_widget.itemClicked.connect(
                lambda item: item.setCheckState(
                    Qt.CheckState.Checked if item.checkState() == Qt.CheckState.Unchecked else Qt.CheckState.Unchecked
                )
            )
        # self.list_widget.setAlternatingRowColors(True)
        self._layout.addRow(self.list_widget)

    def on_selection_changed(self) -> None:
        """Update selection changed information."""
        self.evt_selection_changed.emit()
        if self.allow_toolbar:
            selected = self.get_checked()
            count = self.list_widget.count()
            self.info_label.setText(f"{len(selected)}/{count}")

    def add_item(self, item_text: str) -> None:
        """Add an item to the list in alphabetical order."""
        items = [self.list_widget.item(i).text() for i in range(self.list_widget.count())]
        items.append(item_text)
        checked = [
            self.list_widget.item(i).checkState() == Qt.CheckState.Checked for i in range(self.list_widget.count())
        ]
        checked.append(False)
        if self.allow_sort:
            index = index_natsorted(items)
            items, checked = order_by_index(items, index), order_by_index(checked, index)  # type: ignore[assignment]

        self.list_widget.clear()
        for text, checked in zip(items, checked):
            item = QListWidgetItem(text)
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsUserCheckable)
            item.setCheckState(Qt.CheckState.Unchecked if not checked else Qt.CheckState.Checked)
            self.list_widget.addItem(item)
        self.on_selection_changed()

    def on_filter(self) -> None:
        """Filter list of items in the table."""
        text = self.filter_by.text().lower()
        for index in range(self.list_widget.count()):
            item = self.list_widget.item(index)
            item.setHidden(text not in item.text().lower())

    def on_hide_deselected(self) -> None:
        """Hide deselected items."""
        for i in range(self.list_widget.count()):
            item = self.list_widget.item(i)
            if item.checkState() == Qt.CheckState.Unchecked:
                item.setHidden(True)

    def on_show_all(self) -> None:
        """Show all items."""
        self.filter_by.setText("")
        for i in range(self.list_widget.count()):
            self.list_widget.item(i).setHidden(False)

    def add_items(self, items: list[str], clear: bool = False) -> None:
        """Add multiple items to the list."""
        if clear:
            self.list_widget.clear()
        for item in items:
            self.add_item(item)

    def remove_item(self, item_text: str) -> None:
        """Remove an item from the list."""
        for i in range(self.list_widget.count()):
            if self.list_widget.item(i).text() == item_text:
                self.list_widget.takeItem(i)
                break

    def on_select_all(self) -> None:
        """Select all items in the list."""
        for i in range(self.list_widget.count()):
            item = self.list_widget.item(i)
            if item.isHidden():
                continue
            item.setCheckState(Qt.CheckState.Checked)

    def on_deselect_all(self) -> None:
        """Deselect all items in the list."""
        for i in range(self.list_widget.count()):
            item = self.list_widget.item(i)
            if item.isHidden():
                continue
            item.setCheckState(Qt.CheckState.Unchecked)

    def on_invert_selection(self) -> None:
        """Invert the selection of all items."""
        for i in range(self.list_widget.count()):
            item = self.list_widget.item(i)
            if item.isHidden():
                continue
            item.setCheckState(
                Qt.CheckState.Checked if item.checkState() == Qt.CheckState.Unchecked else Qt.CheckState.Unchecked
            )

    def get_checked(self) -> list[str]:
        """Get checked items."""
        return [
            self.list_widget.item(i).text()
            for i in range(self.list_widget.count())
            if self.list_widget.item(i).checkState() == Qt.CheckState.Checked
        ]

    def set_checked(self, checked: list[str]) -> None:
        """Set checked items."""
        for i in range(self.list_widget.count()):
            item = self.list_widget.item(i)
            current_state = item.checkState()
            item.setCheckState(Qt.CheckState.Checked if item.text() in checked else current_state)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)
    frame.setMinimumSize(600, 600)

    wdg = QtSelectionList(frame)
    wdg.add_item("Item 1")
    wdg.add_item("Item 2")
    wdg.add_item("Item 3")
    ha.addWidget(wdg)

    frame.show()
    sys.exit(app.exec_())



"""Simple dialog to select between two options."""

from __future__ import annotations

import typing as ty

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog, QHBoxLayout, QVBoxLayout, QWidget

import qtextra.helpers as hp

Orientation = ty.Literal["horizontal", "vertical"]


class QtPickOptionBase(QDialog):
    """Select between options."""

    option: ty.Optional[str] = None
    orientation: str | Orientation = "horizontal"

    def __init__(self, parent: QWidget, text: str, options: ty.Dict[str, ty.Any]):
        super().__init__(parent)
        self.setWindowTitle("Select option")

        self.text = text
        self.options = options
        self.responses: dict[str, ty.Any] = {}
        self._setup_ui()

    def _get_layout_widget(self) -> tuple[QWidget | None, QWidget | None]:
        """Get layout widget."""
        raise NotImplementedError("Must implement method")

    def _setup_ui(self) -> None:
        from functools import partial

        area, widget = self._get_layout_widget()
        btn_layout = QHBoxLayout(area) if self.orientation == "horizontal" else QVBoxLayout(area)
        # btn_layout.addStretch(1)
        for label, option in self.options.items():
            btn = hp.make_btn(
                self, label, object_name="pick_option_button", func=partial(self.on_accept, option=option), wrap=True
            )
            btn_layout.addWidget(btn)
            self.responses[btn.text()] = option
        # btn_layout.addStretch(1)

        layout = QVBoxLayout(self)
        layout.addWidget(
            hp.make_label(
                self,
                self.text,
                enable_url=True,
                wrap=True,
                object_name="pick_option_label",
                alignment=Qt.AlignmentFlag.AlignCenter,
            )
        )
        if widget:
            layout.addWidget(widget, stretch=True)
        else:
            layout.addLayout(btn_layout, stretch=True)

    def on_accept(self, option: str) -> None:
        """Set accepted."""
        self.option = option
        return self.accept()

    def reject(self) -> None:
        """Set rejected."""
        self.option = None
        return super().reject()


class QtPickOption(QtPickOptionBase):
    """Select between options."""

    def _get_layout_widget(self) -> tuple[QWidget | None, QWidget | None]:
        """Get layout widget."""
        return None, None


class QtScrollablePickOption(QtPickOptionBase):
    """Select between options."""

    def __init__(
        self,
        parent: QWidget,
        text: str,
        options: dict[str, ty.Any],
        orientation: str | Orientation = "horizontal",
        max_width: int = 500,
    ):
        self.orientation = orientation
        super().__init__(parent, text, options)
        size = self.sizeHint()
        size.setWidth(max(size.width() + 50, max_width))
        size.setHeight(min(400, 40 * len(options) + 70))
        self.setMinimumSize(size)

    def _get_layout_widget(self) -> tuple[QWidget | None, QWidget | None]:
        """Get layout widget."""
        scroll_area, scroll_widget = hp.make_scroll_area(self)
        return scroll_area, scroll_widget


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import apply_style, qapplication

    app = qapplication(False)
    frame = QtScrollablePickOption(
        None,
        "Select an option",
        {
            "Option 1": "option1",
            "Option 2": "option2",
            "Option 3": "option3",
            "Option 4": "option4",
            "Option 5 (dict)": {"test": "1", "name": "option5"},
            "Option 6": "option6",
            "Option 7": "option7",
            **{f"Option {i}": f"option{i}" for i in range(8, 20)},
        },
        orientation="vertical",
    )
    apply_style(frame)

    if frame.exec_():
        print(frame.option)
    sys.exit()



"""Button with progress bar."""

from __future__ import annotations

from qtpy.QtWidgets import QHBoxLayout, QProgressBar, QVBoxLayout, QWidget


class QtActiveProgressBarButton(QWidget):
    """Button with progress bar in a layout."""

    def __init__(self, parent: QWidget | None, text: str = ""):
        super().__init__(parent=parent)

        import qtextra.helpers as hp

        self.active_btn = hp.make_active_btn(self, text)
        self.evt_clicked = self.active_btn.clicked
        self.setText = self.active_btn.setText

        self.progress_bar = QProgressBar(self)
        self.progress_bar.setObjectName("progress_timer")
        self.progress_bar.setTextVisible(False)
        self.progress_bar.setMaximumHeight(10)
        hp.set_sizer_policy(self.progress_bar, h_stretch=True, v_stretch=False)
        hp.set_retain_hidden_size_policy(self.progress_bar)

        self.cancel_btn = hp.make_qta_btn(self, "cancel", average=True)
        self.cancel_btn.setVisible(False)
        self.evt_cancel = self.cancel_btn.clicked

        button_layout = QHBoxLayout()
        button_layout.setContentsMargins(0, 0, 0, 0)
        button_layout.setSpacing(2)
        button_layout.addWidget(self.active_btn, stretch=True)
        button_layout.addWidget(self.cancel_btn)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addLayout(button_layout)
        layout.addWidget(self.progress_bar)
        self.active = False

    @property
    def active(self) -> bool:
        """Update the state of the loading label."""
        return self.active_btn.active

    @active.setter
    def active(self, value: bool) -> None:
        self.active_btn.active = value
        self.cancel_btn.setVisible(value)
        self.progress_bar.setValue(0)
        self.progress_bar.setRange(0, 0)
        self.progress_bar.setVisible(value)

    @property
    def step(self) -> int:
        """Set current step."""
        return self.progress_bar.value()

    @step.setter
    def step(self, value: int) -> None:
        self.progress_bar.setValue(value)
        self.progress_bar.setToolTip(f"Step {value} of {self.progress_bar.maximum()}")

    def setMaximum(self, max_val: int) -> None:
        """Set maximum."""
        self.progress_bar.setMaximum(max_val)
        self.progress_bar.setToolTip(f"Step {self.progress_bar.value()} of {max_val}")

    def setMinimum(self, min_val: int) -> None:
        """Set minimum."""
        self.progress_bar.setMinimum(min_val)
        self.progress_bar.setToolTip(f"Step {self.progress_bar.value()} of {self.progress_bar.maximum()}")

    def setValue(self, value: int) -> None:
        """Set value."""
        self.progress_bar.setValue(value)
        self.progress_bar.setToolTip(f"Step {value} of {self.progress_bar.maximum()}")

    def setRange(self, min_val: int, max_val: int) -> None:
        """Set range."""
        self.progress_bar.setRange(min_val, max_val)
        self.progress_bar.setToolTip(f"Step {self.progress_bar.value()} of {max_val}")


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    def _test1():
        btn1.setValue(btn1.step + 1)
        btn1.active = not btn1.active

    def _test2():
        btn2.setValue(btn2.step + 1)
        btn2.active = not btn2.active

    app, frame, ha = qframe(False)
    frame.setMinimumSize(600, 600)
    btn1 = QtActiveProgressBarButton(frame)
    btn1.evt_clicked.connect(_test1)
    btn1.setText("TEST BUTTON")
    btn1.setRange(0, 100)
    ha.addWidget(btn1)

    btn2 = QtActiveProgressBarButton(frame)
    btn2.evt_clicked.connect(_test2)
    btn2.setText("TEST BUTTON")
    btn2.setRange(0, 100)
    ha.addWidget(btn2)

    frame.show()
    sys.exit(app.exec_())



"""Multi-selection widget."""

from __future__ import annotations

import typing as ty

from qtpy.QtCore import QEvent, QObject, Qt, Signal
from qtpy.QtGui import QAction
from qtpy.QtWidgets import QFormLayout, QWidget

import qtextra.helpers as hp
from qtextra.config import THEMES
from qtextra.utils.table_config import TableConfig
from qtextra.widgets.qt_dialog import QtFramelessPopup
from qtextra.widgets.qt_table_view_check import MultiColumnSingleValueProxyModel, QtCheckableTableView


def filter_selected(current_options: list[str], all_options: list[str]) -> list[str]:
    """Filter selected options."""
    return [option for option in current_options if option in all_options]


def format_options(options: list[str]) -> str:
    """Format options."""
    return "; ".join(options)


def unformat_options(options: str) -> list[str]:
    """Unformat options."""
    return [option.strip() for option in options.split(";") if option.strip()]


class SelectionWidget(QtFramelessPopup):
    """Selection widget."""

    TABLE_CONFIG = (
        TableConfig()  # type: ignore[no-untyped-call]
        .add("", "check", "bool", 25, no_sort=True, hidden=False, sizing="fixed")
        .add("option", "option", "str", 100, sizing="stretch")
    )

    # Signals
    evt_update = Signal(list)
    evt_changed = Signal(list)
    evt_temp_changed = Signal(list)

    cancel_clicked = False
    options: list[str] | None = None
    original_options: list[str] | None = None

    def __init__(self, parent: QWidget, title: str = "Select...", text: str = "", n_max: int = 0, min_width: int = 500):
        self.title = title
        self.text = text
        self.n_max = n_max
        super().__init__(parent)
        self.setMinimumWidth(min_width)
        self.setMinimumHeight(350)
        self.filter_by_option.setFocus()

    def set_options(self, options: list[str], selected_options: list[str] | None = None) -> None:
        """Set options."""
        if not selected_options:
            selected_options = []

        data = []
        selected_options_ = []
        for option in options:
            is_selected = option in selected_options
            data.append([is_selected, option])
            if is_selected:
                selected_options_.append(option)
        self.table.reset_data()
        self.table.add_data(data)
        self.options = selected_options_
        self.original_options = selected_options
        widths = [len(v) for v in options] or [50]
        if widths:
            self.setMinimumWidth(max([self.minimumWidth(), max(widths) * 10]))

    @property
    def selected_options(self) -> list[str]:
        """Return selected options."""
        indices = self.table.get_all_checked()
        self.options = [self.table.get_value(self.TABLE_CONFIG.option, index) for index in indices]
        return self.options

    def accept(self) -> None:
        """Return state."""
        # self.evt_changed.emit(self.selected_options)
        self.evt_update.emit(self.selected_options)
        super().accept()

    def on_cancel(self) -> None:
        """Return state."""
        self.cancel_clicked = True
        self.reject()

    def reject(self) -> None:
        """Return state."""
        if self.cancel_clicked:
            options = self.original_options
        else:
            options = self.selected_options
        self.evt_update.emit(options)
        super().reject()

    def on_check(self, index: int, _state: bool) -> None:
        """Check."""
        indices = self.table.get_all_checked()
        options = [self.table.get_value(self.TABLE_CONFIG.option, index) for index in indices]
        self.evt_temp_changed.emit(options)
        if self.n_max and len(indices) > self.n_max:
            other_index = next(i for i in indices if i != index)
            self.table.set_value(self.TABLE_CONFIG.check, other_index, False)

    def on_show_selected(self) -> None:
        """Show/hide selected."""
        self.table_proxy.setFilterByState(True, self.TABLE_CONFIG.check)

    def on_show_unselected(self) -> None:
        """Show/hide unselected."""
        self.table_proxy.setFilterByState(False, self.TABLE_CONFIG.check)

    def on_show_selected_clear(self) -> None:
        """Show/hide unselected."""
        self.table_proxy.setFilterByState(None, self.TABLE_CONFIG.check)

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QFormLayout:
        """Make panel."""
        self.table = QtCheckableTableView(self, config=self.TABLE_CONFIG, enable_all_check=True, sortable=True)
        self.table.setCornerButtonEnabled(False)
        self.table.evt_checked.connect(self.on_check)
        hp.set_font(self.table, THEMES.get_font_size())
        self.table.setup_model(
            self.TABLE_CONFIG.header, self.TABLE_CONFIG.no_sort_columns, self.TABLE_CONFIG.hidden_columns
        )
        self.table_proxy = MultiColumnSingleValueProxyModel(self)
        self.table_proxy.setSourceModel(self.table.model())
        self.table.model().table_proxy = self.table_proxy
        self.table.setModel(self.table_proxy)

        self.filter_by_option = hp.make_line_edit(
            self,
            placeholder="Filter by value...",
            func_changed=lambda text, col=self.TABLE_CONFIG.option: self.table_proxy.setFilterByColumn(text, col),
        )

        layout = hp.make_form_layout(parent=self, margin=6)

        if self.text:
            layout.addRow(
                hp.make_label(
                    self,
                    self.text,
                    alignment=Qt.AlignmentFlag.AlignHCenter,
                    wrap=True,
                    enable_url=True,
                )
            )
        layout.addRow(self.table)
        layout.addRow(
            hp.make_h_layout(
                hp.make_qta_btn(self, "visible_on", func=self.on_show_selected, tooltip="Only show checked items."),
                hp.make_qta_btn(
                    self, "visible_off", func=self.on_show_unselected, tooltip="Only show unchecked items."
                ),
                hp.make_qta_btn(self, "clear", func=self.on_show_selected_clear, tooltip="Clear checked filter."),
                self.filter_by_option,
                stretch_id=(3,),
                spacing=2,
            )
        )
        layout.addRow(
            hp.make_h_layout(
                hp.make_btn(self, "OK", func=self.accept), hp.make_btn(self, "Cancel", func=self.on_cancel)
            )
        )
        return layout


class QtMultiSelect(QWidget):
    """Multi select widget."""

    editingFinished = Signal()
    textChanged = Signal(str)
    evt_selection_changed = Signal(list)

    def __init__(
        self,
        parent: QWidget,
        allow_clear: bool = False,
        instant_set: bool = False,
        n_max: int = 0,
        hover_opens: bool = False,
    ):
        self.instant_set = instant_set
        self.n_max = n_max
        self.hover_opens = hover_opens

        super().__init__(parent)
        self.options: list[str] = []
        self.selected_options: list[str] = []
        self.text_edit = hp.make_line_edit(self, placeholder="Select...")
        self.text_edit.setReadOnly(True)
        self.text_edit.setClearButtonEnabled(allow_clear)
        self.text_edit.installEventFilter(self)

        self._clear_action = self.text_edit.findChild(QAction)
        if self._clear_action:
            self._clear_action.setEnabled(True)
            self._clear_action.triggered.connect(self.clear_current)

        self._list_action = hp.make_action(
            self, "list", func=self.on_select, tooltip="Click here to select one or more options"
        )
        self.text_edit.addAction(self._list_action, self.text_edit.ActionPosition.LeadingPosition)

        layout = hp.make_h_layout(self.text_edit, stretch_id=0, spacing=0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.text_edit.editingFinished.connect(self.editingFinished.emit)
        self.text_edit.textChanged.connect(self.textChanged.emit)

    def enterEvent(self, event: ty.Any) -> None:
        """Enter event."""
        if self.hover_opens:
            self.text_edit.setFocus()
        super().enterEvent(event)

    def eventFilter(self, obj: QObject, evt: QEvent) -> bool:
        """Event filter."""
        if self.isEnabled() and obj == self.text_edit and evt.type() == QEvent.Type.MouseButtonPress:  # type: ignore
            self.on_select()
            return True
        return super().eventFilter(obj, evt)

    @classmethod
    def from_schema(
        cls: type[QtMultiSelect],
        parent: QWidget,
        description: str = "",
        options: list[str] | None = None,
        value: str = "",
        default: str = "",
        placeholder: str = "Select...",
        func: ty.Callable | ty.Sequence[ty.Callable] | None = None,
        func_changed: ty.Callable | ty.Sequence[ty.Callable] | None = None,
        items: dict[str, ty.Any] | None = None,
        n_max: int = 0,
        **_kwargs: ty.Any,
    ) -> QtMultiSelect:
        """Init."""
        if default:
            value = default
        if value is None:
            value = default
        if items and "enum" in items:
            options = items["enum"]
        if isinstance(value, str):
            values = value.split(";") if value else []
        else:
            values = value

        obj = cls(parent, n_max=n_max)
        obj.text_edit.setPlaceholderText(placeholder)
        obj.options = options or []
        obj.selected_options = values
        obj.text_edit.setText(format_options(obj.selected_options))
        obj.setToolTip(description)
        obj.text_edit.setToolTip(description)
        if func:
            [obj.editingFinished.connect(func_) for func_ in hp._validate_func(func)]
        if func_changed:
            [obj.textChanged.connect(func_) for func_ in hp._validate_func(func_changed)]
        return obj

    def polish(self) -> None:
        """Polish widget."""
        hp.polish_widget(self, self.text_edit)

    def setObjectName(self, name: str) -> None:  # type: ignore
        """Set object name."""
        self.text_edit.setObjectName(name)
        super().setObjectName(name)

    def currentOptions(self) -> list[str]:
        """Return current options."""
        return self.options

    def clear_current(self) -> None:
        """Clear selection."""
        self.selected_options = []
        self.text_edit.setText("")
        self.evt_selection_changed.emit(self.selected_options)

    def clear(self) -> None:
        """Clear selection."""
        self.text_edit.setText("")
        self.selected_options = []
        self.options = []

    def toggle_clear_action(self) -> None:
        """Toggle visibility of clear action."""
        if not self.text_edit.isClearButtonEnabled():
            return
        text = self.text_edit.text()
        self._clear_action.setVisible(text != "")

    def set_options(self, options: list[str], selected_options: list[str] | None = None) -> None:
        """List of options."""
        if selected_options is None:
            selected_options = []
        selected_options = filter_selected(selected_options, options)
        self.options = options
        self.selected_options = selected_options
        self.text_edit.setText(format_options(selected_options))

    def set_selected_options(self, selected_options: list[str]) -> None:
        """List of options."""
        if selected_options is None:
            selected_options = []
        if not isinstance(selected_options, list):
            selected_options = [selected_options]
        selected_options = list(map(str, selected_options))
        if self.options:
            selected_options = filter_selected(selected_options, self.options)
        self.selected_options = selected_options
        self.text_edit.setText(format_options(selected_options))
        self.toggle_clear_action()

    def _set_selected_options(self, selected_options: list[str]) -> None:
        """List of options."""
        if not selected_options:
            selected_options = []
        self.selected_options = selected_options
        self.text_edit.setText(format_options(selected_options))
        # trigger update events
        self.textChanged.emit(format_options(selected_options))
        self.editingFinished.emit()
        self.evt_selection_changed.emit(selected_options)
        self.toggle_clear_action()

    def set_selected_options_temp(self, selected_options: list[str]) -> None:
        """List of options."""
        with hp.qt_signals_blocked(self.text_edit):
            self.set_selected_options(selected_options)
            self.toggle_clear_action()

    def on_select(self) -> None:
        """Select."""
        dlg = SelectionWidget(self, n_max=self.n_max, min_width=self.text_edit.width())
        dlg.set_options(self.options, self.selected_options)
        dlg.filter_by_option.setFocus()
        dlg.evt_temp_changed.connect(self._set_selected_options if self.instant_set else self.set_selected_options_temp)
        dlg.evt_update.connect(self._set_selected_options)
        dlg.show_below_widget(self, x_offset=0, y_offset=0)

    def get_checked(self) -> list[str]:
        """Return list of checked values."""
        return self.selected_options


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)
    frame.setLayout(ha)

    wdg = QtMultiSelect(frame)
    wdg.set_options(["option1", "option2", "option3"], ["option1", "option3"])
    ha.addWidget(wdg)
    wdg = QtMultiSelect(frame, allow_clear=True)
    wdg.set_options(["option1", "option2", "option3"], ["option1", "option3"])
    ha.addWidget(wdg)

    frame.show()
    sys.exit(app.exec_())



"""QtAction."""

import typing as ty

import qtawesome
from qtpy.QtWidgets import QAction

from qtextra.assets import get_icon
from qtextra.config import THEMES


class QtQtaAction(QAction):
    """Icon."""

    _qta_data = None

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        THEMES.evt_theme_icon_changed.connect(self._update_qta)

    def set_qta(self, name: str, **kwargs: ty.Any):
        """Set QtAwesome icon."""
        name, kwargs_ = get_icon(name)
        kwargs.update(kwargs_)
        self._qta_data = (name, kwargs)
        icon = qtawesome.icon(name, **self._qta_data[1], color=THEMES.get_hex_color("icon"))
        self.setIcon(icon)

    def _update_qta(self):
        """Update qta icon."""
        if self._qta_data:
            name, kwargs = self._qta_data
            self.set_qta(name, **kwargs)



import pytest

try:
    from qtextra.dialogs.qt_console import QtConsole
except ImportError:
    QtConsole = None


@pytest.mark.skipif(QtConsole is None, reason="IPython not installed")
def test_console(qtbot):
    """Test creating the console."""
    console = QtConsole()
    qtbot.addWidget(console)

    assert console.kernel_client is not None


@pytest.mark.skipif(QtConsole is None, reason="IPython not installed")
def test_console_user_variables(qtbot):
    """Test creating the console with user variables."""
    console = QtConsole({"var": 3})
    qtbot.addWidget(console)

    assert console.kernel_client is not None
    assert "var" in console.shell.user_ns
    assert console.shell.user_ns["var"] == 3


@pytest.mark.skipif(QtConsole is None, reason="IPython not installed")
def test_multiple_consoles(qtbot):
    """Test creating multiple consoles."""
    console_a = QtConsole({"var_a": 3})
    qtbot.addWidget(console_a)
    console_b = QtConsole({"var_b": 4})
    qtbot.addWidget(console_b)

    assert console_a.kernel_client is not None
    assert console_b.kernel_client is not None
    assert "var_a" in console_a.shell.user_ns
    assert "var_b" in console_a.shell.user_ns



"""Alternative notification."""

from __future__ import annotations

import typing as ty
from contextlib import suppress
from functools import partial

from qtpy.QtCore import QEasingCurve, QRect, QSize, Qt, QThread, QTimer
from qtpy.QtWidgets import (
    QApplication,
    QDialog,
    QFormLayout,
    QHBoxLayout,
    QProgressBar,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)
from superqt import ensure_main_thread
from superqt.utils import qthrottled

import qtextra.helpers as hp
from qtextra.config import EVENTS, THEMES, get_settings
from qtextra.utils.notifications import NOTIFICATION_LEVELS, ErrorNotification, Notification, NotificationSeverity
from qtextra.widgets.qt_button_icon import QtExpandButton
from qtextra.widgets.qt_dialog import QtFramelessPopup, SubWindowBase
from qtextra.widgets.qt_label_icon import QtSeverityLabel

ActionSequence = ty.Sequence[ty.Tuple[str, ty.Callable[[], None]]]


# TODO: This is currently broken since the settings.nottifications settings module is not present


class QtNotification(SubWindowBase):
    """Small popup notification that can contain actions."""

    MAX_OPACITY = 0.9
    FADE_IN_RATE = 220
    FADE_OUT_RATE = 120
    DISMISS_AFTER = 5000
    MIN_WIDTH = 400
    MIN_HEIGHT = 40
    BOTTOM_Y_OFFSET = 0
    MIN_EXPANSION = 18

    def __init__(
        self,
        message: str,
        severity: str | NotificationSeverity = NotificationSeverity.WARNING,
        source: str | None = None,
        actions: ActionSequence = (),
        is_bottom: bool = True,
        expanded: bool = True,
        auto_close: bool = True,
    ):
        super().__init__()
        # disable window decorations
        self.setMinimumWidth(self.MIN_WIDTH)
        self.setMaximumWidth(self.MIN_WIDTH)
        self.setMinimumHeight(self.MIN_HEIGHT)

        self.is_bottom = is_bottom
        # if the notification is not meant to automatically close, set the DISMISS_AFTER value to 0
        if not auto_close:
            self.DISMISS_AFTER = 0

        parent = None
        if parent is not None:
            self.setParent(parent)
        if hasattr(parent, "evt_window_resize"):
            self.parent().evt_window_resize.connect(
                self.move_to_bottom_right if self.is_bottom else self.move_to_top_right
            )
            self.BOTTOM_Y_OFFSET = parent.statusbar.height()

        self.timer = QTimer()
        self.timer_bar = QTimer()

        self.make_ui()
        self.set_notification(message, severity, actions, source)
        if expanded:
            self.setProperty("expanded", str(expanded))
            self.expand_btn.expanded = True if self.is_bottom else False

        EVENTS.evt_notification_dismiss.connect(self.close)

        # setup
        self.adjustSize()
        self.move_to_bottom_right() if self.is_bottom else self.move_to_top_right()

    # noinspection PyAttributeOutsideInit
    def make_ui(self):
        """Make ui.

        Row 1: Icon, text, expand, close
        Row 2: Source, Push buttons
        """
        self.row1_widget = QWidget(self)
        self.severity_icon = QtSeverityLabel(parent=self.row1_widget)
        self.severity_icon.setMaximumWidth(20)
        self.severity_icon.setMinimumHeight(20)

        self.message = hp.make_label(self.row1_widget, wrap=True)
        # self.message = QElidingLabel(self.row1_widget)
        self.message.setWordWrap(True)
        self.message.setMinimumWidth(self.MIN_WIDTH - 200)
        self.message.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.message.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.MinimumExpanding)

        self.expand_btn = QtExpandButton(parent=self.row1_widget)
        self.expand_btn.expanded = True
        self.expand_btn.setToolTip("Click here to expand/contract text")
        self.expand_btn.setCursor(Qt.PointingHandCursor)
        self.expand_btn.setMaximumWidth(20)
        self.expand_btn.setMinimumHeight(20)
        self.expand_btn.clicked.connect(self.toggle_expansion)

        self.settings_btn = hp.make_qta_btn(self.row1_widget, "gear", tooltip="Show notification settings.", small=True)
        self.settings_btn.setCursor(Qt.PointingHandCursor)
        self.settings_btn.setMaximumWidth(20)
        self.settings_btn.setMinimumHeight(20)
        self.settings_btn.clicked.connect(self.settings)

        self.dismiss_btn = hp.make_qta_btn(self.row1_widget, "cross", tooltip="Dismiss this notification.", small=True)
        self.dismiss_btn.setCursor(Qt.PointingHandCursor)
        self.dismiss_btn.setMaximumWidth(20)
        self.dismiss_btn.setMinimumHeight(20)
        self.dismiss_btn.clicked.connect(self.close)

        self._timer_indicator = QProgressBar(self)
        self._timer_indicator.setObjectName("progress_timer")
        self._timer_indicator.setTextVisible(False)

        row_1 = QHBoxLayout(self.row1_widget)
        row_1.addWidget(self.severity_icon, alignment=Qt.AlignmentFlag.AlignTop)
        row_1.addWidget(self.message, stretch=1, alignment=Qt.AlignmentFlag.AlignTop)
        row_1.addWidget(self.expand_btn, alignment=Qt.AlignmentFlag.AlignTop)
        row_1.addWidget(self.settings_btn, alignment=Qt.AlignmentFlag.AlignTop)
        row_1.addWidget(self.dismiss_btn, alignment=Qt.AlignmentFlag.AlignTop)
        row_1.setSpacing(5)

        self.row2_widget = QWidget(self)
        self.row2_widget.setObjectName("actionButtons")
        self.row2_widget.hide()
        self.row2_widget.setMaximumHeight(34)

        self.row2 = QHBoxLayout(self.row2_widget)
        self.row2.addStretch(1)
        self.row2.setContentsMargins(12, 2, 16, 12)

        self.vertical_layout = QVBoxLayout(self)
        self.vertical_layout.setContentsMargins(2, 2, 2, 2)
        self.vertical_layout.setSpacing(2)
        self.vertical_layout.addWidget(self.row1_widget)
        self.vertical_layout.addWidget(self.row2_widget)
        self.vertical_layout.addStretch(1)
        self.vertical_layout.addWidget(self._timer_indicator)
        self.resize(self.MIN_WIDTH, self.MIN_HEIGHT)

    def set_notification(
        self,
        message: str,
        severity: NotificationSeverity = NotificationSeverity.INFO,
        actions: ActionSequence = None,
        source: str | None = None,
    ):
        """Set message."""
        self.message.setText(message)
        self.message.setToolTip(message)
        self.setup_buttons(actions)
        self.severity_icon.severity = str(severity)
        self.expand_btn.setVisible(self.sizeHint().height() > self.height())

    def slide_in(self):
        """Run animation that fades in the dialog with a slight slide up."""
        super().slide_in()
        if get_settings().notification.auto_expand:
            self.opacity_anim.finished.connect(self.auto_expand)

    def auto_expand(self):
        """Toggle expansion."""
        self.expand()

    def toggle_expansion(self):
        """Toggle the expanded state of the notification frame."""
        self.contract() if self.is_expanded else self.expand()
        self.timer.stop()

    def expand(self):
        """Expanded the notification so that the full message is visible."""
        curr = self.geometry()
        self.geom_anim.setDuration(100)
        self.geom_anim.setStartValue(curr)
        new_height = self.sizeHint().height()
        if new_height < curr.height():
            # new height would shift notification down, ensure some expansion
            new_height = curr.height() + self.MIN_EXPANSION
        delta = new_height - curr.height()
        self.geom_anim.setEndValue(
            QRect(
                curr.x(),
                curr.y() - delta if self.is_bottom else curr.y(),
                curr.width(),
                new_height,
            )
        )
        self.geom_anim.setEasingCurve(QEasingCurve.OutQuad)
        self.geom_anim.start()
        self.setProperty("expanded", True if self.is_bottom else False)
        self.expand_btn.expanded = False if self.is_bottom else True

    def contract(self):
        """Contract notification to a single elided line of the message."""
        geom = self.geometry()
        self.geom_anim.setDuration(100)
        self.geom_anim.setStartValue(geom)
        dlt = geom.height() - self.minimumHeight()
        self.geom_anim.setEndValue(
            QRect(
                geom.x(),
                geom.y() + dlt if self.is_bottom else geom.y(),
                geom.width(),
                geom.height() - dlt,
            )
        )
        self.geom_anim.setEasingCurve(QEasingCurve.OutQuad)
        self.geom_anim.start()
        self.setProperty("expanded", False if self.is_bottom else True)
        self.expand_btn.expanded = True if self.is_bottom else False

    def show(self):
        """Show the message with a fade and slight slide in from the bottom."""

        def _update_timer_indicator():
            with suppress(RuntimeError):
                self._timer_indicator.setValue(self.timer.remainingTime() / self.DISMISS_AFTER * 100)

        super().show()
        self.slide_in()
        if self.DISMISS_AFTER > 0:
            self.timer.setInterval(get_settings().notification.dismiss_time)
            self.timer.setSingleShot(True)
            self.timer.timeout.connect(self.close)
            self.timer.start()

            self.timer_bar.setInterval(50)
            self.timer_bar.setSingleShot(False)
            self.timer_bar.timeout.connect(_update_timer_indicator)
            self.timer_bar.start()

    def close(self):
        """Fade out then close."""
        try:
            self.opacity_anim.setDuration(self.FADE_OUT_RATE)
            self.opacity_anim.setStartValue(self.MAX_OPACITY)
            self.opacity_anim.setEndValue(0)
            self.opacity_anim.start()
            self.opacity_anim.finished.connect(super().close)
        except RuntimeError:
            pass

    def mouseMoveEvent(self, event):
        """On hover, stop the self-destruct timer."""
        self.timer.stop()
        self.timer_bar.stop()
        self._timer_indicator.setVisible(False)
        return super().mouseMoveEvent(event)

    def setup_buttons(self, actions: ActionSequence = ()):
        """Add buttons to the dialog.

        Parameters
        ----------
        actions : tuple, optional
            A sequence of 2-tuples, where each tuple is a string and a
            callable. Each tuple will be used to create button in the dialog,
            where the text on the button is determine by the first item in the
            tuple, and a callback function to call when the button is pressed
            is the second item in the tuple. by default ()
        """
        if actions is None:
            return

        if isinstance(actions, dict):
            actions = list(actions.items())

        for text, callback in actions:
            btn = hp.make_btn(None, text)

            def call_back_with_self(callback, self):
                """We need a higher order function this to capture the reference to self."""

                def _inner():
                    return callback(self)

                return _inner

            btn.clicked.connect(call_back_with_self(callback, self))
            btn.clicked.connect(self.close)
            self.row2.addWidget(btn)
        if actions:
            self.row2_widget.show()
            self.setMinimumHeight(self.row2_widget.maximumHeight() + self.minimumHeight())

    def sizeHint(self):
        """Return the size required to show the entire message."""
        return QSize(
            super().sizeHint().width(),
            self.row2_widget.height() + self.message.sizeHint().height() + 30,
        )

    @property
    def is_expanded(self) -> bool:
        """Checks whether text is expanded."""
        return self.property("expanded") if self.is_bottom else not self.property("expanded")

    def settings(self):
        """Show settings popup."""
        self.timer.stop()
        SettingsPopup(self.parent() or self).show()

    @classmethod
    def from_notification(cls, notification: Notification) -> QtNotification:
        """From notification."""
        actions = notification.actions
        if isinstance(notification, ErrorNotification):
            actions = tuple(notification.actions) + (
                # ("Copy to clipboard", partial(copy_to_clipboard, notification)),
                ("Report on GitHub", partial(show_report, notification)),
                ("View Traceback", partial(show_tb, notification)),
            )

        return cls(
            message=notification.message,
            severity=notification.severity,
            source=notification.source,
            actions=actions,
            # is_bottom=notification.severity != NotificationSeverity.SUCCESS,
            #             expanded=notification.severity == NotificationSeverity.SUCCESS,
            auto_close=notification.auto_close,
        )

    @classmethod
    @qthrottled(timeout=250)
    @ensure_main_thread
    def show_notification(cls, notification: Notification):
        """Show notification."""
        if notification.severity >= get_settings().notification.level:
            application_instance = QApplication.instance()
            if application_instance:
                # Check if this is running from a thread
                if application_instance.thread() != QThread.currentThread():
                    EVENTS.evt_notification.emit(notification)
                    return
            cls.from_notification(notification).show()


class SettingsPopup(QtFramelessPopup):
    """Popup window to control extra settings."""

    def __init__(self, parent):
        super().__init__(parent)
        self.setMinimumWidth(300)

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QFormLayout:
        """Make panel."""
        self.dismiss_time = hp.make_labelled_slider(
            self,
            minimum=1000,
            maximum=10000,
            step_size=500,
            tooltip="Specify the amount of time before the popup is automatically dismissed. Time in milliseconds.",
        )
        self.dismiss_time.setValue(get_settings().notification.dismiss_time)
        self.dismiss_time.valueChanged.connect(self.on_apply)

        self.auto_expand = hp.make_checkbox(self, tooltip="Auto-expand notification if the message is too long.")
        self.auto_expand.setChecked(get_settings().notification.auto_expand)
        self.auto_expand.stateChanged.connect(self.on_apply)

        self.level = hp.make_combobox(self, tooltip="Specify what kind of notifications you wish to see.")
        hp.set_combobox_data(self.level, NOTIFICATION_LEVELS, get_settings().notification.level)
        self.level.currentTextChanged.connect(self.on_apply)

        main_layout = hp.make_form_layout()
        main_layout.addRow(
            hp.make_label(self, "Notification Settings", alignment=Qt.AlignmentFlag.AlignHCenter, bold=True)
        )
        main_layout.addRow(hp.make_label(self, "Auto-dismiss time (ms)"), self.dismiss_time)
        main_layout.addRow(hp.make_label(self, "Auto-expand"), self.auto_expand)
        main_layout.addRow(hp.make_label(self, "Notification level"), self.level)
        return main_layout

    def on_apply(self):
        """Update configuration."""
        settings = get_settings()
        settings.notification.dismiss_time = self.dismiss_time.value()
        settings.notification.auto_expand = self.auto_expand.isChecked()
        settings.notification.level = self.level.currentData()


def show_tb(notification: ErrorNotification, parent):
    """Show traceback."""
    tb_dialog = QDialog(parent=parent.parent())
    tb_dialog.setModal(True)
    # this is about the minimum width to not get re-wrap and the minimum height to not have scrollbar
    tb_dialog.resize(650, 270)
    tb_dialog.setLayout(QVBoxLayout())

    text = hp.make_text_edit(None, "")
    text.setHtml(notification.as_html())
    text.setReadOnly(True)
    debug_btn = hp.make_btn(None, "Enter Debugger")

    def _enter_debug_mode():
        debug_btn.setText("Now Debugging. Please quit debugger in console to continue")
        _debug_tb(notification.exception.__traceback__)
        debug_btn.setText("Enter Debugger")

    debug_btn.clicked.connect(_enter_debug_mode)
    tb_dialog.layout().addWidget(text)
    tb_dialog.layout().addWidget(debug_btn, 0, Qt.AlignmentFlag.AlignRight)
    tb_dialog.show()


def copy_to_clipboard(notification: ErrorNotification, parent):
    """Copy traceback to clipboard."""
    from qtextra.utils.utilities import get_system_info
    from qtextra.widgets.qt_button_clipboard import copy_text_to_clipboard

    html = "<b>More information</b><br><br>" + get_system_info(as_html=True) + "<br><br><b>Exception</b><br>"
    exception = notification.as_html()
    html += f"``{exception}``"

    text = hp.make_text_edit(None, "")
    text.setHtml(html)

    copy_text_to_clipboard(text.toMarkdown())
    text.deleteLater()


def show_report(notification: ErrorNotification, parent):
    """Report traceback."""
    import webbrowser

    from qtextra import __issue_url__
    from qtextra.utils.utilities import get_system_info
    from qtextra.widgets.qt_button_clipboard import copy_text_to_clipboard

    def _go_to_github():
        copy_text_to_clipboard(text.toMarkdown())
        webbrowser.open(__issue_url__)
        github_btn.setText("Copied report to clipboard!")

    def _copy_to_clipboard():
        copy_text_to_clipboard(text.toMarkdown())
        copy_btn.setText("Copied!")
        hp.add_flash_animation(text, color=THEMES.get_hex_color("foreground"), duration=500)

    report_dialog = QDialog(parent=parent.parent())
    report_dialog.setModal(True)
    # this is about the minimum width to not get re-wrap and the minimum height to not have scrollbar
    report_dialog.resize(650, 270)
    report_dialog.setLayout(QVBoxLayout())

    html = get_system_info(as_html=True) + "<br><br><b>Exception</b><br>"
    exception = notification.as_html()
    html += f"``{exception}``"

    text = hp.make_text_edit(None, "")
    text.setHtml(html)
    text.setReadOnly(True)

    copy_btn = hp.make_btn(None, "Copy to clipboard")
    copy_btn.clicked.connect(_copy_to_clipboard)
    github_btn = hp.make_btn(None, "Go to GitHub")
    github_btn.clicked.connect(_go_to_github)

    btn_layout = QHBoxLayout()
    btn_layout.addStretch(True)
    btn_layout.addWidget(copy_btn, 0, Qt.AlignmentFlag.AlignRight)
    btn_layout.addWidget(github_btn, 0, Qt.AlignmentFlag.AlignRight)

    report_dialog.layout().addWidget(text)
    report_dialog.layout().addLayout(btn_layout)
    report_dialog.show()


def _debug_tb(tb):
    import pdb

    from qtextra.helpers import event_hook_removed

    QApplication.processEvents()
    QApplication.processEvents()
    with event_hook_removed():
        print("Entering debugger. Type 'q' to return to the app.\n")
        pdb.post_mortem(tb)
        print("\nDebugging finished.  App active again.")


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys
        from random import choice

        from qtextra.config import THEMES
        from qtextra.utils.dev import qframe

        def _popup_notif(severity=None):
            if not isinstance(severity, NotificationSeverity):
                severity = choice(list(NotificationSeverity))
            if severity == NotificationSeverity.ERROR:
                notif = ErrorNotification(ValueError("This is going to be quite a long exception\n" * 4))
            else:
                notif = Notification(
                    message=f"This is a test message: {severity!s}\n" * choice(range(5)), severity=severity
                )
            pop = QtNotification.from_notification(notif)
            THEMES.set_theme_stylesheet(pop)
            pop.show()

        def _popup_error():
            _popup_notif(NotificationSeverity.ERROR)

        def _warning():
            notif = Notification(
                message="Some requirements were not met - see what is not quite right.",
                severity=NotificationSeverity.WARNING,
                actions=(("Open Requirements Dialog", print),),
            )
            pop = QtNotification.from_notification(notif)
            THEMES.set_theme_stylesheet(pop)
            pop.show()

        app, frame, ha = qframe(False)
        frame.setMinimumSize(600, 600)

        btn2 = hp.make_btn(frame, "Create random notification")
        btn2.clicked.connect(_popup_notif)
        ha.addWidget(btn2)
        btn2 = hp.make_btn(frame, "Create error notification")
        btn2.clicked.connect(_popup_error)
        ha.addWidget(btn2)
        btn2 = hp.make_btn(frame, "Create warning notification")
        btn2.clicked.connect(_warning)
        ha.addWidget(btn2)

        ha.addWidget(btn2)
        frame.show()
        sys.exit(app.exec_())

    _main()



from pathlib import Path

import appdirs

APP_NAME = "qtextra"
APP_AUTHOR = False

USER_DATA_DIR = Path(appdirs.user_data_dir(APP_NAME, APP_AUTHOR))

USER_CACHE_DIR = USER_DATA_DIR / "Cache"
USER_CACHE_DIR.mkdir(parents=True, exist_ok=True)
USER_CONFIG_DIR = USER_DATA_DIR / "Configs"
USER_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
USER_LOG_DIR = USER_DATA_DIR / "Logs"
USER_LOG_DIR.mkdir(parents=True, exist_ok=True)
USER_THEME_DIR = USER_DATA_DIR / "Themes"
USER_THEME_DIR.mkdir(parents=True, exist_ok=True)

CONFIG_FILENAME = "settings.yaml"



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from psygnal import Signal as PySignal
from qtpy.QtCore import QLineF, QPoint, QPointF, QRectF, Qt, Signal
from qtpy.QtGui import QBrush, QColor, QFont, QMouseEvent, QPainterPath, QPen, QResizeEvent, QShowEvent, QWheelEvent
from qtpy.QtWidgets import (
    QGraphicsItem,
    QGraphicsPathItem,
    QGraphicsRectItem,
    QGraphicsSceneMouseEvent,
    QGraphicsSimpleTextItem,
    QMenu,
    QWidgetAction,
)

from qtextra.utils.color import get_text_color
from qtextra.widgets._qt_graphics import QBaseGraphicsScene, QBaseGraphicsView
from qtextra.widgets.qt_line_edit import QDoubleLineEdit

if TYPE_CHECKING:
    import numpy.typing as npt


def quick_min_max(
    arr: npt.NDArray,
    down_sample_to: int = 1048576,
) -> tuple[float, float]:
    """Quickly compute the min and max of an array."""
    if arr.dtype.kind == "b":
        return 0.0, 1.0
    down_sample_factor = arr.size / down_sample_to
    if down_sample_factor <= 1.0:
        return arr.min(), arr.max()
    stride = int(np.ceil(down_sample_factor))
    arr_ref = arr[::stride]
    return arr_ref.min(), arr_ref.max()


class QHistogramView(QBaseGraphicsView):
    """Graphics view for displaying histograms and setting contrast limits."""

    evt_clim_changed = Signal(tuple)

    _pos_drag_start: QPoint | None = None
    _pos_drag_prev: QPoint | None = None

    def __init__(self):
        super().__init__()
        self._hist_items = [self.addItem(QHistogramItem())]
        self._line_low = self.addItem(QClimLineItem(0))
        self._line_low.evt_value_changed.connect(self._on_clim_changed)
        self._line_high = self.addItem(QClimLineItem(1))
        self._line_high.evt_value_changed.connect(self._on_clim_changed)
        self._view_range: tuple[float, float] = (0.0, 1.0)
        self._minmax = (0.0, 1.0)  # limits of the movable range

    def _on_clim_changed(self):
        clim = self.clim()
        if self._view_range is not None:
            v0, v1 = self._view_range
            x0, x1 = self.clim()
            if x0 < v0 or x1 > v1:
                self._view_range = clim
                self.update()
        self.evt_clim_changed.emit(clim)

    def clim(self) -> tuple[float, float]:
        """The current contrast limits."""
        return tuple(sorted([self._line_low.value(), self._line_high.value()]))

    def set_clim(self, clim: tuple[float, float]):
        """Set the contrast limits."""
        self._line_low.setValue(max(clim[0], self._minmax[0]))
        self._line_high.setValue(min(clim[1], self._minmax[1]))

    def set_minmax(self, minmax: tuple[float, float]):
        """Set the minimum and maximum values that the contrast limits can take."""
        self._minmax = minmax
        self._line_low.setRange(*minmax)
        self._line_high.setRange(*minmax)

    def set_histogram_from_array(
        self,
        arr: npt.NDArray,
        clim: tuple[float, float],
        is_rgb: bool = False,
        color: QColor = QColor(100, 100, 100),
        minmax: tuple[float, float] | None = None,
    ):
        """Set the histogram for an array."""
        # coerce the number of histogram items
        n_hist = 3 if is_rgb else 1
        for _ in range(n_hist, len(self._hist_items)):
            self.scene().removeItem(self._hist_items[-1])
            self._hist_items.pop()
        for _ in range(len(self._hist_items), n_hist):
            self._hist_items.append(self.addItem(QHistogramItem()))

        if is_rgb:
            brushes = [
                QBrush(QColor(255, 0, 0, 128)),
                QBrush(QColor(0, 255, 0, 128)),
                QBrush(QColor(0, 0, 255, 255)),
            ]  # RGB
            for i, (item, brush) in enumerate(zip(self._hist_items, brushes)):
                item.with_brush(brush).set_histogram_from_array(arr[:, :, i])
        else:
            brushes = [QBrush(color)]
            self._hist_items[0].with_brush(brushes[0]).set_histogram_from_array(arr)

        if not minmax:
            if arr.dtype.kind in "ui":
                minmax = (np.iinfo(arr.dtype).min, np.iinfo(arr.dtype).max)
            elif arr.dtype.kind == "b":
                minmax = (0, 1)
            else:
                minmax = clim
        self.set_minmax(minmax)
        self.set_clim(clim)
        if self._view_range is None:
            self._view_range = clim

    def setValueFormat(self, fmt: str):
        """Set the value format for the contrast limits."""
        self._line_low.setValueFormat(fmt)
        self._line_high.setValueFormat(fmt)

    def viewRect(self, width: float | None = None) -> QRectF:
        """The current view range as a QRectF."""
        x0, x1 = self._view_range
        if width is None:
            width = x1 - x0
        return QRectF(x0 - width * 0.03, 0, width * 1.06, 1)

    def set_view_range(self, x0: float, x1: float):
        """Set the view range."""
        if (x0, x1) == self._view_range:
            return
        self._view_range = (x0, x1)
        self.fitInView(self.viewRect(), Qt.AspectRatioMode.IgnoreAspectRatio)

    def resizeEvent(self, event: QResizeEvent):
        super().resizeEvent(event)
        self.fitInView(self.viewRect(), Qt.AspectRatioMode.IgnoreAspectRatio)

    def showEvent(self, event: QShowEvent):
        super().showEvent(event)
        x0, x1 = self.clim()
        self.fitInView(self.viewRect(x1 - x0), Qt.AspectRatioMode.IgnoreAspectRatio)
        self._line_low.setValue(self._line_low.value())
        self._line_high.setValue(self._line_high.value())

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        rect = self._line_low.boundingRect().united(self._line_high.boundingRect())
        for hist in self._hist_items:
            rect = rect.united(hist.boundingRect())
        x0, x1 = rect.left(), rect.right()
        self.fitInView(self.viewRect(x1 - x0), Qt.AspectRatioMode.IgnoreAspectRatio)
        self._view_range = (x0, x1)

    def wheelEvent(self, event: QWheelEvent):
        delta = event.angleDelta().y()
        if delta > 0:
            factor = 1.1
        else:
            factor = 1 / 1.1
        x0, x1 = self._view_range
        xcursor = self.mapToScene(event.pos()).x()
        x0 = max((x0 - xcursor) / factor + xcursor, self._minmax[0])
        x1 = min((x1 - xcursor) / factor + xcursor, self._minmax[1])
        self.set_view_range(x0, x1)

    def mousePressEvent(self, event: QMouseEvent):
        if event.button() == Qt.MouseButton.LeftButton:
            self._pos_drag_start = event.pos()
            self._pos_drag_prev = self._pos_drag_start
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QMouseEvent):
        if self.scene().grabSource():
            return super().mouseMoveEvent(event)
        if event.buttons() & Qt.MouseButton.LeftButton:
            pos = event.pos()
            if self._pos_drag_prev is not None:
                delta = self.mapToScene(pos) - self.mapToScene(self._pos_drag_prev)
                delta = delta.x()
                x0, x1 = self._view_range
                x0 -= delta
                x1 -= delta
                self.set_view_range(x0, x1)
            self._pos_drag_prev = pos
        return super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent):
        self._pos_drag_start = None
        self._pos_drag_prev = None
        self.scene().setGrabSource(None)
        return super().mouseReleaseEvent(event)


class QClimLineItem(QGraphicsRectItem):
    """The line item for one of the contrast limits."""

    # NOTE: To properly set the bounding rect, we need to inherit from QGraphicsRectItem
    # with updated boundingRect method.
    evt_value_changed = PySignal(float)
    _Y_LOW = -1
    _Y_HIGH = 2
    _WIDTH_NORMAL = 3
    _WIDTH_HOVER = 6

    def __init__(self, x: float):
        super().__init__()
        self._color = QColor(255, 0, 0, 150)
        pen = QPen(self._color, self._WIDTH_NORMAL)
        pen.setCosmetic(True)
        self._qpen = pen
        self.setZValue(1000)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable)
        self._is_dragging = False
        self._range = (-float("inf"), float("inf"))
        self._value = x
        self._value_fmt = ".1f"
        self.setCursor(Qt.CursorShape.SizeHorCursor)
        self._value_label = QGraphicsSimpleTextItem()
        self._value_label.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations)
        self._value_label.setFont(QFont("Arial", 12))
        self.setAcceptHoverEvents(True)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent):
        if event.button() == Qt.MouseButton.LeftButton:
            self._is_dragging = True
            self.scene().setGrabSource(self)
        elif event.buttons() & Qt.MouseButton.RightButton:
            self.scene().setGrabSource(self)
            menu = QClimMenu(self.scene().views()[0], self)
            menu._edit.setFocus()
            menu.exec(event.screenPos())
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent):
        if event.buttons() & Qt.MouseButton.LeftButton:
            if self._is_dragging:
                self._drag_event(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent):
        self._is_dragging = False
        self.scene().setGrabSource(None)
        self.setValue(event.pos().x())
        return super().mouseReleaseEvent(event)

    def hoverEnterEvent(self, event):
        self._qpen.setWidthF(self._WIDTH_HOVER)
        self._show_value_label()
        self.update()
        return super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        self._qpen.setWidthF(self._WIDTH_NORMAL)
        self._value_label.hide()
        self.update()
        return super().hoverLeaveEvent(event)

    def _show_value_label(self):
        txt = format(self.value(), self._value_fmt)
        self._value_label.setText(txt)
        vp = self.scene().views()[0].viewport()
        background_color = vp.palette().color(vp.backgroundRole())
        self._value_label.setBrush(QBrush(get_text_color(background_color)))
        text_width = self._value_label.boundingRect().width()
        pos = QPointF(self.value(), 0)
        if pos.x() + text_width / self._x_scale() > self._range[1]:
            pos.setX(pos.x() - (text_width + 4) / self._x_scale())
        else:
            pos.setX(pos.x() + 4 / self._x_scale())
        self._value_label.setPos(self.mapToScene(pos))
        if self._value_label.scene() is None:
            # prevent scene movement during adding the label
            rect = self.scene().sceneRect()
            self.scene().addItem(self._value_label)
            self.scene().setSceneRect(rect)
        self._value_label.show()

    def _drag_event(self, event: QGraphicsSceneMouseEvent):
        self.setValue(event.pos().x())
        self._show_value_label()
        if scene := self.scene():
            scene.update()

    def setValueFormat(self, fmt: str):
        self._value_fmt = fmt

    def paint(self, painter, option, widget):
        painter.setPen(self._qpen)
        start = QPointF(self._value, self._Y_LOW)
        end = QPointF(self._value, self._Y_HIGH)
        line = QLineF(start, end)
        painter.drawLine(line)

    def value(self) -> float:
        """The x value of the line (the contrast limit)."""
        return self._value

    def setValue(self, x: float):
        """Set the x value of the line (the contrast limit)."""
        old_bbox = self.boundingRect()
        old_value = self._value
        new_value = min(max(x, self._range[0]), self._range[1])
        self._value = new_value
        new_bbox = self.boundingRect()
        self.setRect(new_bbox)
        if new_value != old_value:
            self.evt_value_changed.emit(self._value)
        if scene := self.scene():
            scene.update(self.mapRectToScene(old_bbox.united(new_bbox)))

    def setRange(self, low: float, high: float):
        """Set the min/max range of the line x value."""
        self._range = (low, high)
        if not low <= self.value() <= high:
            self.setValue(self.value())

    def scene(self) -> QBaseGraphicsScene:
        return super().scene()

    def _x_scale(self) -> float:
        return self.view().transform().m11()

    def view(self) -> QHistogramView:
        return self.scene().views()[0]

    def boundingRect(self):
        w = 10.0 / self._x_scale()
        x = self.value()
        return QRectF(x - w / 2, self._Y_LOW, w, self._Y_HIGH - self._Y_LOW)


class QHistogramItem(QGraphicsPathItem):
    """The histogram item."""

    def __init__(self):
        super().__init__()
        self._hist_path = QPainterPath()
        self._hist_brush = QBrush(QColor(100, 100, 100))
        self.setPen(QPen(QColor(0, 0, 0, 0)))

    def with_brush(self, brush: QBrush) -> QHistogramItem:
        """Set the brush for the histogram."""
        self._hist_brush = brush
        return self

    def set_histogram_from_array(self, arr: npt.NDArray) -> tuple[float, float]:
        """Set the histogram from an array."""
        _min, _max = quick_min_max(arr)
        if arr.dtype in ("uint8", "uint8"):
            _nbin = 64
        else:
            _nbin = 256
        _nbin = min(_nbin, int(np.prod(arr.shape[-2:])) // 2)
        # draw histogram
        if arr.dtype.kind == "b":
            edges = np.array([0, 0.5, 1])
            frac_true = np.sum(arr) / arr.size
            hist = np.array([1 - frac_true, frac_true])
        elif _max > _min:
            arr = arr.clip(_min, _max)
            if arr.dtype.kind in "ui" and _max - _min < _nbin:
                # bin number is excessive
                _nbin = int(_max - _min)
                normed = (arr - _min).astype(np.uint8)
            else:
                normed = ((arr - _min) / (_max - _min) * _nbin).astype(np.uint8)
            hist = np.bincount(normed.ravel(), minlength=_nbin)
            hist = hist / hist.max()
            edges = np.linspace(_min, _max, _nbin + 1)
        else:
            edges = np.array([_min, _max])
            hist = np.zeros(1)
        _path = QPainterPath()
        self.setBrush(self._hist_brush)
        _path.moveTo(edges[0], 1)
        for e0, e1, h in zip(edges[:-1], edges[1:], hist):
            _path.lineTo(e0, 1 - h)
            _path.lineTo(e1, 1 - h)
        _path.lineTo(edges[-1], 1)
        _path.closeSubpath()
        self.setPath(_path)
        self.update()


class QClimMenu(QMenu):
    """Menu."""

    def __init__(self, parent: QHistogramView, item: QClimLineItem):
        super().__init__(parent)
        self._hist_view = parent
        self._item = item
        self._edit = QDoubleLineEdit()
        self._edit.setText(format(item.value(), item._value_fmt))
        self._edit.valueChanged.connect(self._on_value_changed)
        self._edit.editingFinished.connect(self._on_value_set)
        widget_action = QWidgetAction(self)
        widget_action.setDefaultWidget(self._edit)
        self.addAction(widget_action)

    def _on_value_changed(self):
        """Value changed."""
        value = self._edit.text()
        if value:
            value = float(value)
            self._item.setValue(value)

    def _on_value_set(self):
        value = self._edit.text()
        if value:
            value = float(value)
            self._item.setValue(value)
            # update min/max
            if value < self._hist_view._minmax[0]:
                self._hist_view.set_minmax((value, self._hist_view._minmax[1]))
            elif value > self._hist_view._minmax[1]:
                self._hist_view.set_minmax((self._hist_view._minmax[0], value))
            # update view range
            v0, v1 = self._hist_view._view_range
            if value < v0:
                self._hist_view.set_view_range(value, v1)
            elif value > v1:
                self._hist_view.set_view_range(v0, value)
        self.close()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(horz=False)
    frame.setMinimumSize(600, 600)
    hist = QHistogramView()
    hist.set_histogram_from_array(np.random.randint(0, 65332, 100), (0, 65332), minmax=(0, 65332))
    ha.addWidget(hist)

    hist = QHistogramView()
    hist.set_histogram_from_array(np.random.randint(0, 255, (100, 100, 3)), (0, 255), is_rgb=True, minmax=(0, 255))
    ha.addWidget(hist)

    frame.show()
    sys.exit(app.exec_())



"""tool tip widget."""

from __future__ import annotations

import typing as ty
from enum import Enum

from qtpy.QtCore import QEvent, QObject, QPoint, QPointF, QPropertyAnimation, Qt, QTimer
from qtpy.QtGui import QColor, QIcon, QImage, QPainter, QPainterPath, QPixmap, QPolygonF
from qtpy.QtWidgets import QGraphicsDropShadowEffect, QHBoxLayout, QWidget

import qtextra.helpers as hp
from qtextra.config import THEMES
from qtextra.widgets.qt_popout import PopoutView, PopoutViewBase


class TipPosition(Enum):
    """tool tip tail position."""

    TOP = 0
    BOTTOM = 1
    LEFT = 2
    RIGHT = 3
    TOP_LEFT = 4
    TOP_RIGHT = 5
    BOTTOM_LEFT = 6
    BOTTOM_RIGHT = 7
    LEFT_TOP = 8
    LEFT_BOTTOM = 9
    RIGHT_TOP = 10
    RIGHT_BOTTOM = 11
    NONE = 12


class ImagePosition(Enum):
    """Image position."""

    TOP = 0
    BOTTOM = 1
    LEFT = 2
    RIGHT = 3


class QtToolTipView(PopoutView):
    """tool tip view."""

    def __init__(
        self,
        title: str,
        content: str,
        icon: ty.Union[QIcon, str] = None,
        image: ty.Union[str, QPixmap, QImage] = None,
        is_closable: bool = True,
        tail_position: TipPosition = TipPosition.BOTTOM,
        parent=None,
    ):
        self.manager = ToolTipManager.make(tail_position)
        self.hBoxLayout = QHBoxLayout()
        self.hBoxLayout.setContentsMargins(0, 0, 0, 0)
        super().__init__(title, content, icon, image, is_closable, parent)

    def _adjustImage(self):
        if self.manager._get_image_position() in [ImagePosition.TOP, ImagePosition.BOTTOM]:
            return super()._adjustImage()

        h = self.vBoxLayout.sizeHint().height() - 2
        self.imageLabel.scaledToHeight(h)

    def _addImageToLayout(self):
        self.imageLabel.setHidden(self.imageLabel.isNull())
        pos = self.manager._get_image_position()

        if pos == ImagePosition.TOP:
            self.imageLabel.setBorderRadius(8, 8, 0, 0)
            self.vBoxLayout.insertWidget(0, self.imageLabel)
        elif pos == ImagePosition.BOTTOM:
            self.imageLabel.setBorderRadius(0, 0, 8, 8)
            self.vBoxLayout.addWidget(self.imageLabel)
        elif pos == ImagePosition.LEFT:
            self.vBoxLayout.removeItem(self.vBoxLayout.itemAt(0))
            self.hBoxLayout.addLayout(self.viewLayout)
            self.vBoxLayout.addLayout(self.hBoxLayout)

            self.imageLabel.setBorderRadius(8, 0, 8, 0)
            self.hBoxLayout.insertWidget(0, self.imageLabel)
        elif pos == ImagePosition.RIGHT:
            self.vBoxLayout.removeItem(self.vBoxLayout.itemAt(0))
            self.hBoxLayout.addLayout(self.viewLayout)
            self.vBoxLayout.addLayout(self.hBoxLayout)

            self.imageLabel.setBorderRadius(0, 8, 0, 8)
            self.hBoxLayout.addWidget(self.imageLabel)

    def paintEvent(self, e):
        pass


class ToolTipBubble(QWidget):
    """tool tip bubble."""

    def __init__(self, view: PopoutViewBase, tail_position: TipPosition = TipPosition.BOTTOM, parent=None):
        super().__init__(parent=parent)
        self.manager = ToolTipManager.make(tail_position)
        self.hBoxLayout = QHBoxLayout(self)
        self.view = view

        self.manager._update_layout(self)
        self.hBoxLayout.addWidget(self.view)

    def setView(self, view: QWidget):
        self.hBoxLayout.removeWidget(self.view)
        self.view.deleteLater()
        self.view = view
        self.hBoxLayout.addWidget(view)

    def paintEvent(self, e):
        painter = QPainter(self)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing)

        painter.setBrush(QColor(40, 40, 40) if THEMES.is_dark else QColor(248, 248, 248))
        painter.setPen(QColor(23, 23, 23) if THEMES.is_dark else QColor(0, 0, 0, 17))

        self.manager.draw(self, painter)


class QtToolTip(QWidget):
    """tool tip."""

    def __init__(
        self,
        view: PopoutViewBase,
        target: QWidget,
        duration: int = 1000,
        tail_position: TipPosition = TipPosition.BOTTOM,
        parent=None,
        delete_on_close: bool = True,
    ):
        """
        Parameters
        ----------
        target: QWidget
            the target widget to show tip

        view: PopoutViewBase
            tool tip view

        duration: int
            the time for tool tip to display in milliseconds. If duration is less than zero,
            tool tip will never disappear.

        tail_position: TipPosition
            the position of bubble tail

        parent: QWidget
            parent widget

        delete_on_close: bool
            whether delete flyout automatically when flyout is closed
        """
        super().__init__(parent=parent)
        self.target = target
        self.duration = duration
        self.isDeleteOnClose = delete_on_close
        self.manager = ToolTipManager.make(tail_position)

        self.hBoxLayout = QHBoxLayout(self)
        self.opacity_animation = QPropertyAnimation(self, b"windowOpacity", self)

        self.bubble = ToolTipBubble(view, tail_position, self)

        self.hBoxLayout.setContentsMargins(15, 8, 15, 20)
        self.hBoxLayout.addWidget(self.bubble)
        self.setShadowEffect()

        # set style
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setWindowFlags(Qt.WindowType.Tool | Qt.WindowType.FramelessWindowHint)

        if parent and parent.window():
            parent.window().installEventFilter(self)

    def setShadowEffect(self, blurRadius=35, offset=(0, 8)):
        """Add shadow to dialog."""
        color = QColor(0, 0, 0, 80 if THEMES.is_dark else 30)
        self.shadowEffect = QGraphicsDropShadowEffect(self.bubble)
        self.shadowEffect.setBlurRadius(blurRadius)
        self.shadowEffect.setOffset(*offset)
        self.shadowEffect.setColor(color)
        self.bubble.setGraphicsEffect(None)
        self.bubble.setGraphicsEffect(self.shadowEffect)

    def _fadeOut(self):
        """Fade out."""
        self.opacity_animation.setDuration(167)
        self.opacity_animation.setStartValue(1)
        self.opacity_animation.setEndValue(0)
        self.opacity_animation.finished.connect(self.close)
        self.opacity_animation.start()

    def showEvent(self, e):
        if self.duration >= 0:
            QTimer.singleShot(self.duration, self._fadeOut)

        self.move(self.manager._get_position(self))
        self.adjustSize()
        self.opacity_animation.setDuration(167)
        self.opacity_animation.setStartValue(0)
        self.opacity_animation.setEndValue(1)
        self.opacity_animation.start()
        super().showEvent(e)

    def closeEvent(self, e):
        if self.isDeleteOnClose:
            self.deleteLater()

        super().closeEvent(e)

    def eventFilter(self, obj, e: QEvent):
        if self.parent() and obj is self.parent().window():
            if e.type() in [QEvent.Type.Resize, QEvent.Type.WindowStateChange, QEvent.Type.Move]:
                self.move(self.manager._get_position(self))

        return super().eventFilter(obj, e)

    def addWidget(self, widget: QWidget, stretch=0, align=Qt.AlignmentFlag.AlignLeft):
        """Add widget to tool tip."""
        self.view.addSpacing(8)
        self.view.addWidget(widget, stretch, align)

    @property
    def view(self):
        return self.bubble.view

    def setView(self, view):
        self.bubble.setView(view)

    @classmethod
    def _init(
        cls,
        view: PopoutViewBase,
        target: QWidget,
        duration=1000,
        tail_position=TipPosition.BOTTOM,
        parent=None,
        delete_on_close=True,
    ):
        """
        Parameters
        ----------
        view: PopoutViewBase
            tool tip view

        target: QWidget
            the target widget to show tip

        duration: int
            the time for tool tip to display in milliseconds. If duration is less than zero,
            tool tip will never disappear.

        tail_position: TipPosition
            the position of bubble tail

        parent: QWidget
            parent widget

        delete_on_close: bool
            whether delete flyout automatically when flyout is closed
        """
        w = cls(view, target, duration, tail_position, parent, delete_on_close)
        w.show()
        return w

    @classmethod
    def init(
        cls,
        target: QWidget,
        title: str,
        content: str,
        icon: ty.Union[QIcon, str] | None = None,
        image: ty.Union[str, QPixmap, QImage] | None = None,
        is_closable: bool = True,
        duration: int = 1000,
        tail_position=TipPosition.BOTTOM,
        parent=None,
        delete_on_close=True,
    ):
        """
        Parameters
        ----------
        target: QWidget
            the target widget to show tip

        title: str
            the title of tool tip

        content: str
            the content of tool tip

        icon:  QIcon | str
            the icon of tool tip

        image: str | QPixmap | QImage
            the image of tool tip

        is_closable: bool
            whether to show the close button

        duration: int
            the time for tool tip to display in milliseconds. If duration is less than zero,
            tool tip will never disappear.
        parent: QWidget
            parent widget
        tail_position: TipPosition
            the position of bubble tail
        delete_on_close: bool
            whether delete flyout automatically when flyout is closed
        """
        view = QtToolTipView(title, content, icon, image, is_closable, tail_position)
        w = cls._init(view, target, duration, tail_position, parent, delete_on_close)
        view.evt_closed.connect(w.close)
        return w


class PopupToolTip(QtToolTip):
    """Pop up tool tip."""

    def __init__(
        self,
        view: PopoutViewBase,
        target: QWidget,
        duration=1000,
        tail_position=TipPosition.BOTTOM,
        parent=None,
        delete_on_close=True,
    ):
        super().__init__(view, target, duration, tail_position, parent, delete_on_close)
        self.setWindowFlags(
            Qt.WindowType.Popup | Qt.WindowType.FramelessWindowHint | Qt.WindowType.NoDropShadowWindowHint
        )


class ToolTipManager(QObject):
    """tool tip manager."""

    def __init__(self):
        super().__init__()

    def _update_layout(self, tip: ToolTipBubble):
        """Manage the layout of tip."""
        tip.hBoxLayout.setContentsMargins(0, 0, 0, 0)

    def _get_image_position(self):
        return ImagePosition.TOP

    def _get_position(self, tip: QtToolTip) -> QPoint:
        pos = self._pos(tip)
        rect = hp.get_current_screen_geometry()
        x = max(rect.left(), min(pos.x(), rect.right() - tip.width() - 4))
        y = max(rect.top(), min(pos.y(), rect.bottom() - tip.height() - 4))
        return QPoint(x, y)

    def draw(self, tip: ToolTipBubble, painter: QPainter):
        """Draw the shape of bubble."""
        rect = tip.rect().adjusted(1, 1, -1, -1)
        painter.drawRoundedRect(rect, 8, 8)

    def _pos(self, tip: QtToolTip):
        """Return the poisition of tip."""
        return tip.pos()

    @staticmethod
    def make(position: TipPosition):
        """Mask tool tip manager according to the display position."""
        managers = {
            TipPosition.TOP: TopTailToolTipManager,
            TipPosition.BOTTOM: BottomTailToolTipManager,
            TipPosition.LEFT: LeftTailToolTipManager,
            TipPosition.RIGHT: RightTailToolTipManager,
            TipPosition.TOP_RIGHT: TopRightTailTeachingTipManager,
            TipPosition.BOTTOM_RIGHT: BottomRightTailTeachingTipManager,
            TipPosition.TOP_LEFT: TopLeftTailTeachingTipManager,
            TipPosition.BOTTOM_LEFT: BottomLeftTailTeachingTipManager,
            TipPosition.LEFT_TOP: LeftTopTailTeachingTipManager,
            TipPosition.LEFT_BOTTOM: LeftBottomTailTeachingTipManager,
            TipPosition.RIGHT_TOP: RightTopTailTeachingTipManager,
            TipPosition.RIGHT_BOTTOM: RightBottomTailTeachingTipManager,
            TipPosition.NONE: ToolTipManager,
        }

        if position not in managers:
            raise ValueError(f"`{position}` is an invalid tool tip position.")

        return managers[position]()


class TopTailToolTipManager(ToolTipManager):
    """Top tail tool tip manager."""

    def _update_layout(self, tip):
        tip.hBoxLayout.setContentsMargins(0, 8, 0, 0)

    def _get_image_position(self):
        return ImagePosition.BOTTOM

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pt = tip.hBoxLayout.contentsMargins().top()

        path = QPainterPath()
        path.addRoundedRect(1, pt, w - 2, h - pt - 1, 8, 8)
        path.addPolygon(QPolygonF([QPointF(w / 2 - 7, pt), QPointF(w / 2, 1), QPointF(w / 2 + 7, pt)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        pos = target.mapToGlobal(QPoint(0, target.height()))
        x = pos.x() + target.width() // 2 - tip.sizeHint().width() // 2
        y = pos.y() - tip.layout().contentsMargins().top()
        return QPoint(x, y)


class BottomTailToolTipManager(ToolTipManager):
    """Bottom tail tool tip manager."""

    def _update_layout(self, tip):
        tip.hBoxLayout.setContentsMargins(0, 0, 0, 8)

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pb = tip.hBoxLayout.contentsMargins().bottom()

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - 2, h - pb - 1, 8, 8)
        path.addPolygon(QPolygonF([QPointF(w / 2 - 7, h - pb), QPointF(w / 2, h - 1), QPointF(w / 2 + 7, h - pb)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        pos = target.mapToGlobal(QPoint())
        x = pos.x() + target.width() // 2 - tip.sizeHint().width() // 2
        y = pos.y() - tip.sizeHint().height() + tip.layout().contentsMargins().bottom()
        return QPoint(x, y)


class LeftTailToolTipManager(ToolTipManager):
    """Left tail tool tip manager."""

    def _update_layout(self, tip):
        tip.hBoxLayout.setContentsMargins(8, 0, 0, 0)

    def _get_image_position(self):
        return ImagePosition.RIGHT

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pl = 8

        path = QPainterPath()
        path.addRoundedRect(pl, 1, w - pl - 2, h - 2, 8, 8)
        path.addPolygon(QPolygonF([QPointF(pl, h / 2 - 7), QPointF(1, h / 2), QPointF(pl, h / 2 + 7)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(target.width(), 0))
        x = pos.x() - m.left()
        y = pos.y() - tip.view.sizeHint().height() // 2 + target.height() // 2 - m.top()
        return QPoint(x, y)


class RightTailToolTipManager(ToolTipManager):
    """Left tail tool tip manager."""

    def _update_layout(self, tip):
        tip.hBoxLayout.setContentsMargins(0, 0, 8, 0)

    def _get_image_position(self):
        return ImagePosition.LEFT

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pr = 8

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - pr - 1, h - 2, 8, 8)
        path.addPolygon(QPolygonF([QPointF(w - pr, h / 2 - 7), QPointF(w - 1, h / 2), QPointF(w - pr, h / 2 + 7)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(0, 0))
        x = pos.x() - tip.sizeHint().width() + m.right()
        y = pos.y() - tip.view.sizeHint().height() // 2 + target.height() // 2 - m.top()
        return QPoint(x, y)


class TopLeftTailTeachingTipManager(TopTailToolTipManager):
    """Top left tail tool tip manager."""

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pt = tip.hBoxLayout.contentsMargins().top()

        path = QPainterPath()
        path.addRoundedRect(1, pt, w - 2, h - pt - 1, 8, 8)
        path.addPolygon(QPolygonF([QPointF(20, pt), QPointF(27, 1), QPointF(34, pt)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        pos = target.mapToGlobal(QPoint(0, target.height()))
        x = pos.x() - tip.layout().contentsMargins().left()
        y = pos.y() - tip.layout().contentsMargins().top()
        return QPoint(x, y)


class TopRightTailTeachingTipManager(TopTailToolTipManager):
    """Top right tail tool tip manager."""

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pt = tip.hBoxLayout.contentsMargins().top()

        path = QPainterPath()
        path.addRoundedRect(1, pt, w - 2, h - pt - 1, 8, 8)
        path.addPolygon(QPolygonF([QPointF(w - 20, pt), QPointF(w - 27, 1), QPointF(w - 34, pt)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        pos = target.mapToGlobal(QPoint(target.width(), target.height()))
        x = pos.x() - tip.sizeHint().width() + tip.layout().contentsMargins().left()
        y = pos.y() - tip.layout().contentsMargins().top()
        return QPoint(x, y)


class BottomLeftTailTeachingTipManager(BottomTailToolTipManager):
    """Bottom left tail tool tip manager."""

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pb = tip.hBoxLayout.contentsMargins().bottom()

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - 2, h - pb - 1, 8, 8)
        path.addPolygon(QPolygonF([QPointF(20, h - pb), QPointF(27, h - 1), QPointF(34, h - pb)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        pos = target.mapToGlobal(QPoint())
        x = pos.x() - tip.layout().contentsMargins().left()
        y = pos.y() - tip.sizeHint().height() + tip.layout().contentsMargins().bottom()
        return QPoint(x, y)


class BottomRightTailTeachingTipManager(BottomTailToolTipManager):
    """Bottom right tail tool tip manager."""

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pb = tip.hBoxLayout.contentsMargins().bottom()

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - 2, h - pb - 1, 8, 8)
        path.addPolygon(QPolygonF([QPointF(w - 20, h - pb), QPointF(w - 27, h - 1), QPointF(w - 34, h - pb)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        pos = target.mapToGlobal(QPoint(target.width(), 0))
        x = pos.x() - tip.sizeHint().width() + tip.layout().contentsMargins().left()
        y = pos.y() - tip.sizeHint().height() + tip.layout().contentsMargins().bottom()
        return QPoint(x, y)


class LeftTopTailTeachingTipManager(LeftTailToolTipManager):
    """Left top tail tool tip manager."""

    def _get_image_position(self):
        return ImagePosition.BOTTOM

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pl = 8

        path = QPainterPath()
        path.addRoundedRect(pl, 1, w - pl - 2, h - 2, 8, 8)
        path.addPolygon(QPolygonF([QPointF(pl, 10), QPointF(1, 17), QPointF(pl, 24)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(target.width(), 0))
        x = pos.x() - m.left()
        y = pos.y() - m.top()
        return QPoint(x, y)


class LeftBottomTailTeachingTipManager(LeftTailToolTipManager):
    """Left bottom tail tool tip manager."""

    def _get_image_position(self):
        return ImagePosition.TOP

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pl = 9

        path = QPainterPath()
        path.addRoundedRect(pl, 1, w - pl - 1, h - 2, 8, 8)
        path.addPolygon(QPolygonF([QPointF(pl, h - 10), QPointF(1, h - 17), QPointF(pl, h - 24)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(target.width(), target.height()))
        x = pos.x() - m.left()
        y = pos.y() - tip.sizeHint().height() + m.bottom()
        return QPoint(x, y)


class RightTopTailTeachingTipManager(RightTailToolTipManager):
    """Right top tail tool tip manager."""

    def _get_image_position(self):
        return ImagePosition.BOTTOM

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pr = 8

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - pr - 1, h - 2, 8, 8)
        path.addPolygon(QPolygonF([QPointF(w - pr, 10), QPointF(w - 1, 17), QPointF(w - pr, 24)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(0, 0))
        x = pos.x() - tip.sizeHint().width() + m.right()
        y = pos.y() - m.top()
        return QPoint(x, y)


class RightBottomTailTeachingTipManager(RightTailToolTipManager):
    """Right bottom tail tool tip manager."""

    def _get_image_position(self):
        return ImagePosition.TOP

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pr = 8

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - pr - 1, h - 2, 8, 8)
        path.addPolygon(QPolygonF([QPointF(w - pr, h - 10), QPointF(w - 1, h - 17), QPointF(w - pr, h - 24)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QtToolTip):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(0, target.height()))
        x = pos.x() - tip.sizeHint().width() + m.right()
        y = pos.y() - tip.sizeHint().height() + m.bottom()
        return QPoint(x, y)


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys
        from random import choice

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe()

        def _popup():
            QtToolTip.init(
                btn,
                title="Title",
                content="Here is some text that should be displayed below the title",
                icon="success",
                parent=frame,
                tail_position=choice(list(TipPosition)),
                is_closable=True,
                duration=3000,
            )

        btn = hp.make_btn(frame, "Show Popout", func=_popup)

        ha.addWidget(btn)
        frame.show()
        sys.exit(app.exec_())

    _main()  # type: ignore[no-untyped-call]



"""Image label."""

from __future__ import annotations

from qtpy.QtCore import QRectF, QSize, Qt, Signal
from qtpy.QtGui import QColor, QImage, QImageReader, QMovie, QPainter, QPainterPath, QPixmap, QResizeEvent, QWheelEvent
from qtpy.QtWidgets import QGraphicsPixmapItem, QGraphicsScene, QGraphicsView, QLabel, QWidget

from qtextra.config import THEMES


class QPixmapLabel(QLabel):
    """Image label."""

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.pixmap_width: int = 1
        self.pixmap_height: int = 1

    def setImage(self, image: str | QPixmap | QImage | None = None):
        """Set the image of label."""
        self.image = image or QImage()

        if isinstance(image, str):
            reader = QImageReader(image)
            if reader.supportsAnimation():
                self.setMovie(QMovie(image))
            else:
                self.image = reader.read()
        elif isinstance(image, QPixmap):
            self.image = image.toImage()

        self.setFixedSize(self.image.size())
        self.update()

    def setPixmap(self, pm) -> None:
        """Set Pixmap."""
        self.pixmap_width = pm.width()
        self.pixmap_height = pm.height()
        self.update_margins()
        super().setPixmap(pm)

    def resizeEvent(self, a0) -> None:
        """Resize event."""
        self.update_margins()
        super().resizeEvent(a0)

    def update_margins(self):
        if self.pixmap() is None:
            return
        pixmapWidth = self.pixmap().width()
        pixmap_height = self.pixmap().height()
        if pixmapWidth <= 0 or pixmap_height <= 0:
            return
        w, h = self.width(), self.height()
        if w <= 0 or h <= 0:
            return

        if w * pixmap_height > h * pixmapWidth:
            m = int((w - (pixmapWidth * h / pixmap_height)) / 2)
            self.setContentsMargins(m, 0, m, 0)
        else:
            m = int((h - (pixmap_height * w / pixmapWidth)) / 2)
            self.setContentsMargins(0, m, 0, m)


class ImageViewer(QGraphicsView):
    """Simple image viewer widget."""

    def __init__(self, image_path=None, parent=None):
        super().__init__(parent)

        # Set up the scene
        self.scene = QGraphicsScene(self)
        self.setScene(self.scene)

        # Add a pixmap item
        self.pixmap_item = QGraphicsPixmapItem()
        self.scene.addItem(self.pixmap_item)

        # Enable panning
        self.setDragMode(QGraphicsView.DragMode.ScrollHandDrag)

        # Track zoom level and adjust zoom factor dynamically
        self.zoom_level = 0
        self.base_zoom_factor = 1.1  # Base zoom factor, will be adjusted dynamically

        # Set up smooth transformation for better quality zoom
        self.setRenderHints(
            self.renderHints() | QPainter.RenderHint.SmoothPixmapTransform | QPainter.RenderHint.Antialiasing
        )

        # Load the initial image if provided
        if image_path:
            self.set_image(image_path)

    def set_image(self, image_path: str):
        """Set or change the image displayed in the viewer."""
        pixmap = QPixmap(image_path)
        if pixmap.isNull():
            print(f"Error: Could not load image from {image_path}")
            return

        # Update the pixmap in the scene
        self.pixmap_item.setPixmap(pixmap)

        # Reset zoom and fit the view to the new image
        self.reset_zoom()
        self.fitInView(self.pixmap_item, Qt.AspectRatioMode.KeepAspectRatio)

    def wheelEvent(self, event: QWheelEvent):
        """Handle zoom in/out with mouse wheel."""
        # Determine the mouse position relative to the scene
        mouse_scene_pos = self.mapToScene(event.position().toPoint())

        # Determine zoom in or out
        zoom_in = event.angleDelta().y() > 0

        if zoom_in:
            self.zoom(1, mouse_scene_pos)
        else:
            self.zoom(-1, mouse_scene_pos)

    def mouseDoubleClickEvent(self, event):
        """Reset zoom level on double-click."""
        self.zoom_level = 0
        self.fitInView(self.pixmap_item, Qt.AspectRatioMode.KeepAspectRatio)

    def zoom(self, direction, center_point):
        """Adjust zoom level and apply scaling centered around a point."""
        if direction > 0:
            factor = self.base_zoom_factor
            self.zoom_level += 1
        elif direction < 0 and self.zoom_level > 0:
            factor = 1 / self.base_zoom_factor
            self.zoom_level -= 1
        else:
            return  # Prevent over-zooming

        # Center the zoom on the mouse pointer
        self.setTransformationAnchor(QGraphicsView.NoAnchor)
        self.setResizeAnchor(QGraphicsView.NoAnchor)

        old_center = self.mapToScene(self.viewport().rect().center())
        self.scale(factor, factor)
        new_center = center_point
        offset = new_center - old_center
        self.centerOn(self.mapToScene(self.viewport().rect().center()) + offset)

    def reset_zoom(self):
        """Reset the zoom to its original state and fit the image to the view."""
        self.resetTransform()
        self.zoom_level = 0
        self.fitInView(self.pixmap_item, Qt.AspectRatioMode.KeepAspectRatio)


class PixmapLabel(QLabel):
    """Label for displaying images."""

    _path = None
    _pixmap = None

    def __init__(self):
        super().__init__()

    def set_image(self, path: str) -> None:
        """Set image from path."""
        if self._path == str(path):
            return
        self._path = str(path)
        self.set_pixmap(QPixmap(str(path)))

    def set_pixmap(self, pm: QPixmap) -> None:
        """Set Pixmap."""
        self._pixmap = pm
        self.setPixmap(pm)

    def setPixmap(self, pm: QPixmap) -> None:
        """Set Pixmap."""
        super().setPixmap(self._resize_pixmap())

    def resizeEvent(self, a0: QResizeEvent) -> None:
        """Resize event."""
        pixmap = self._resize_pixmap()
        if pixmap:
            self.setPixmap(pixmap)
        super().resizeEvent(a0)

    def _resize_pixmap(self):
        if self._pixmap is None or self.pixmap() is None:
            return

        width = self.width()
        height = self.height()
        pixmap = self._pixmap.scaled(width, height, Qt.AspectRatioMode.KeepAspectRatio)
        return pixmap


class QImageLabel(QLabel):
    """Image label."""

    evt_clicked = Signal()

    def __init__(self, parent: QWidget = None):
        super().__init__(parent)
        self.image = QImage()
        self.setBorderRadius(0, 0, 0, 0)
        self._postInit()

    def _postInit(self):
        pass

    def _onFrameChanged(self, index: int):
        self.image = self.movie().currentImage()
        self.update()

    def setBorderRadius(self, topLeft: int, topRight: int, bottomLeft: int, bottomRight: int):
        """Set the border radius of image."""
        self._topLeftRadius = topLeft
        self._topRightRadius = topRight
        self._bottomLeftRadius = bottomLeft
        self._bottomRightRadius = bottomRight
        self.update()

    def setImage(self, image: str | QPixmap | QImage | None = None):
        """Set the image of label."""
        self.image = image or QImage()

        if isinstance(image, str):
            reader = QImageReader(image)
            if reader.supportsAnimation():
                self.setMovie(QMovie(image))
            else:
                self.image = reader.read()
        elif isinstance(image, QPixmap):
            self.image = image.toImage()

        self.setFixedSize(self.image.size())
        self.update()

    def scaledToWidth(self, width: int):
        if self.isNull():
            return

        h = int(width / self.image.width() * self.image.height())
        self.setFixedSize(width, h)

        if self.movie():
            self.movie().setScaledSize(QSize(width, h))

    def scaledToHeight(self, height: int):
        """Scale the image to a specific height."""
        if self.isNull():
            return

        w = int(height / self.image.height() * self.image.width())
        self.setFixedSize(w, height)

        if self.movie():
            self.movie().setScaledSize(QSize(w, height))

    def setScaledSize(self, size: QSize):
        if self.isNull():
            return

        self.setFixedSize(size)

        if self.movie():
            self.movie().setScaledSize(size)

    def isNull(self):
        return self.image.isNull()

    def mouseReleaseEvent(self, e):
        super().mouseReleaseEvent(e)
        self.evt_clicked.emit()

    def setPixmap(self, pixmap: QPixmap):
        self.setImage(pixmap)

    def pixmap(self) -> QPixmap:
        return QPixmap.fromImage(self.image)

    def setMovie(self, movie: QMovie):
        super().setMovie(movie)
        self.movie().start()
        self.image = self.movie().currentImage()
        self.movie().frameChanged.connect(self._onFrameChanged)

    def paintEvent(self, e):
        if self.isNull():
            return

        painter = QPainter(self)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing)

        path = QPainterPath()
        w, h = self.width(), self.height()

        # top line
        path.moveTo(self.topLeftRadius, 0)
        path.lineTo(w - self.topRightRadius, 0)

        # top right arc
        d = self.topRightRadius * 2
        path.arcTo(w - d, 0, d, d, 90, -90)

        # right line
        path.lineTo(w, h - self.bottomRightRadius)

        # bottom right arc
        d = self.bottomRightRadius * 2
        path.arcTo(w - d, h - d, d, d, 0, -90)

        # bottom line
        path.lineTo(self.bottomLeftRadius, h)

        # bottom left arc
        d = self.bottomLeftRadius * 2
        path.arcTo(0, h - d, d, d, -90, -90)

        # left line
        path.lineTo(0, self.topLeftRadius)

        # top left arc
        d = self.topLeftRadius * 2
        path.arcTo(0, 0, d, d, -180, -90)

        # draw image
        image = self.image.scaled(
            self.size() * self.devicePixelRatioF(),
            Qt.AspectRatioMode.IgnoreAspectRatio,
            Qt.TransformationMode.SmoothTransformation,
        )

        painter.setPen(Qt.PenStyle.NoPen)
        painter.setClipPath(path)
        painter.drawImage(self.rect(), image)

    @property
    def topLeftRadius(self):
        return self._topLeftRadius

    @topLeftRadius.setter
    def topLeftRadius(self, radius: int):
        self.setBorderRadius(radius, self.topRightRadius, self.bottomLeftRadius, self.bottomRightRadius)

    @property
    def topRightRadius(self):
        return self._topRightRadius

    @topRightRadius.setter
    def topRightRadius(self, radius: int):
        self.setBorderRadius(self.topLeftRadius, radius, self.bottomLeftRadius, self.bottomRightRadius)

    @property
    def bottomLeftRadius(self):
        return self._bottomLeftRadius

    @bottomLeftRadius.setter
    def bottomLeftRadius(self, radius: int):
        self.setBorderRadius(self.topLeftRadius, self.topRightRadius, radius, self.bottomRightRadius)

    @property
    def bottomRightRadius(self):
        return self._bottomRightRadius

    @bottomRightRadius.setter
    def bottomRightRadius(self, radius: int):
        self.setBorderRadius(self.topLeftRadius, self.topRightRadius, self.bottomLeftRadius, radius)


class AvatarWidget(QImageLabel):
    """Avatar widget."""

    def _postInit(self):
        self.setRadius(48)
        self.lightBackgroundColor = QColor(0, 0, 0, 50)
        self.darkBackgroundColor = QColor(255, 255, 255, 50)

    def getRadius(self):
        return self._radius

    def setRadius(self, radius: int):
        from qtextra.helpers import set_font

        self._radius = radius
        set_font(self, radius)
        self.setFixedSize(2 * radius, 2 * radius)
        self.update()

    def setBackgroundColor(self, light: QColor, dark: QColor):
        self.lightBackgroundColor = QColor(light)
        self.darkBackgroundColor = QColor(light)
        self.update()

    def paintEvent(self, e):
        painter = QPainter(self)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing)

        if not self.isNull():
            self._drawImageAvatar(painter)
        else:
            self._drawTextAvatar(painter)

    def _drawImageAvatar(self, painter: QPainter):
        # center crop image
        image = self.image.scaled(
            self.size() * self.devicePixelRatioF(),
            Qt.AspectRatioMode.KeepAspectRatioByExpanding,
            Qt.TransformationMode.SmoothTransformation,
        )  # type: QImage

        iw, ih = image.width(), image.height()
        d = self.getRadius() * 2 * self.devicePixelRatioF()
        x, y = (iw - d) / 2, (ih - d) / 2
        image = image.copy(int(x), int(y), int(d), int(d))

        # draw image
        path = QPainterPath()
        path.addEllipse(QRectF(self.rect()))

        painter.setPen(Qt.NoPen)
        painter.setClipPath(path)
        painter.drawImage(self.rect(), image)

    def _drawTextAvatar(self, painter: QPainter):
        if not self.text():
            return

        painter.setBrush(self.darkBackgroundColor if THEMES.is_dark else self.lightBackgroundColor)
        painter.setPen(Qt.NoPen)
        painter.drawEllipse(QRectF(self.rect()))

        painter.setFont(self.font())
        painter.setPen(Qt.white if THEMES.is_dark else Qt.black)
        painter.drawText(self.rect(), Qt.AlignCenter, self.text()[0].upper())

    radius = property(int, getRadius, setRadius)



"""Toggle group."""

from __future__ import annotations

import typing as ty

from qtpy.QtCore import Signal
from qtpy.QtWidgets import QFrame, QWidget


class QtToggleGroup(QFrame):
    """Widget for toggle group."""

    _old_value: ty.Any = None

    evt_changed = Signal(object)

    def __init__(
        self,
        parent: QWidget | None,
        options: list[str],
        value: str = "",
        tooltip: str = "",
        orientation: str = "horizontal",
        exclusive: bool = True,
        multiline: bool = False,
    ):
        super().__init__(parent)
        self.setFrameShape(QFrame.Shape.Box)
        self.setLineWidth(1)
        self.setMinimumHeight(26)

        import qtextra.helpers as hp

        layout, self.button_group = hp.make_toggle_group(
            self,
            *options,
            checked_label=value,
            tooltip=tooltip,
            func=self._on_changed,
            orientation=orientation,
            exclusive=exclusive,
            multiline=multiline,
        )
        layout.setContentsMargins(2, 2, 2, 2)
        layout.setSpacing(1)
        self.setLayout(layout)

    def _on_changed(self, _: ty.Any) -> None:
        if self._old_value != self.value:
            self._old_value = self.value
            self.evt_changed.emit(self.value)

    @property
    def buttons(self) -> list[QWidget]:
        """Buttons."""
        return self.button_group.buttons()

    @property
    def checked_buttons(self) -> ty.Any:
        """Button."""
        value = []
        for button in self.button_group.buttons():
            if button.isChecked():
                value.append(button)
        return value

    @property
    def value(self) -> str | list[str] | None:
        """Get value."""
        buttons = self.checked_buttons
        value = [button.text() for button in buttons]
        if value and self.button_group.exclusive():
            return value[0]
        return value

    @value.setter
    def value(self, value: str | list[str]) -> None:
        """Set value."""
        if not isinstance(value, list):
            value = [value]
        for button in self.button_group.buttons():
            if button.text() in value:
                button.setChecked(True)

    @classmethod
    def from_schema(
        cls: type[QtToggleGroup],
        parent: QWidget | None,
        options: list[str] | None = None,
        value: str | list[str] | None = None,
        tooltip: str = "",
        orientation: str = "horizontal",
        default: str | list[str] = "",
        description: str = "",
        items: dict[str, ty.Any] | None = None,
        func: ty.Callable | None = None,
        exclusive: bool = True,
        multiline: bool = False,
        **kwargs: dict,
    ) -> QtToggleGroup:
        """From schema."""
        import qtextra.helpers as hp

        if default:
            value = default
        if value is None:
            value = default
        if description and not tooltip:
            tooltip = description
        if items and "enum" in items:
            options = items["enum"]
        if "enum" in kwargs:
            options = kwargs["enum"]
        widget = cls(
            parent,
            options=options,
            value=value,
            tooltip=tooltip,
            orientation=orientation,
            exclusive=exclusive,
            multiline=multiline,
        )
        if func:
            [widget.evt_changed.connect(func_) for func_ in hp._validate_func(func)]
        return widget


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)
    frame.setLayout(ha)

    wdg = QtToggleGroup.from_schema(None, [str(i) for i in range(15)], "1", func=print)
    ha.addWidget(wdg)

    wdg = QtToggleGroup.from_schema(None, [str(i) for i in range(15)], ["1", "5"], exclusive=False, func=print)
    ha.addWidget(wdg)

    wdg = QtToggleGroup.from_schema(
        None, [str(i) for i in range(15)], ["1", "5"], exclusive=False, func=print, orientation="flow"
    )
    ha.addWidget(wdg)

    wdg = QtToggleGroup.from_schema(None, [str(i) for i in range(15)], "1", func=print)
    wdg.setObjectName("error")
    ha.addWidget(wdg)

    wdg = QtToggleGroup.from_schema(None, [str(i) for i in range(15)], "1", func=print)
    wdg.setObjectName("warning")
    ha.addWidget(wdg)

    frame.show()
    sys.exit(app.exec_())



"""Scrollable label."""

from __future__ import annotations

import typing as ty

from koyo.typing import PathLike
from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QLabel, QScrollArea, QWidget


class QtScrollableLabel(QScrollArea):
    """Scrollable label."""

    evt_clicked = Signal()

    def __init__(
        self,
        parent: ty.Optional[QWidget] = None,
        text: ty.Optional[str] = None,
        image_path: ty.Optional[PathLike] = None,
        wrap: bool = False,
    ):
        super().__init__(parent)

        self.label = QLabel()
        self.label.setWordWrap(wrap)
        self.label.setAlignment(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop)

        if text:
            self.setText(text)
        elif image_path:
            self.set_image(image_path)

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setWidgetResizable(True)
        self.setWidget(self.label)
        self.setContentsMargins(0, 0, 0, 0)

    def setText(self, text: str) -> None:
        """Set label text."""
        self.label.setText(text)

    def set_image(self, image_path: PathLike) -> None:
        """Set image."""
        pixmap = QPixmap(str(image_path))
        self.label.setPixmap(pixmap)
        self.label.adjustSize()

    def clear(self) -> None:
        """Clear label."""
        self.label.clear()

    def mousePressEvent(self, event) -> None:
        """Mouse press event."""
        self.evt_clicked.emit()
        return super().mousePressEvent(event)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)
    frame.setMinimumSize(600, 600)

    ha.addWidget(QLabel("Scrollable label without wrap"))
    wdg = QtScrollableLabel(text="This is a long text that will make the QLabel scrollable. " * 20)
    ha.addWidget(wdg)
    ha.addWidget(QLabel("Scrollable label with wrap"))
    wdg = QtScrollableLabel(text="This is a long text that will make the QLabel scrollable. " * 20, wrap=True)
    ha.addWidget(wdg)

    frame.show()
    sys.exit(app.exec_())



import re
import typing as ty
from contextlib import suppress

from koyo.system import IS_WIN

from qtextra.typing import Callback

if ty.TYPE_CHECKING:
    from qtextra.queue.task import Task

COLORS = {
    "error": "#ff121e",
    "warning": "#ff693c",
    "check_warning": "#ff693c",
    "success": "#1ed75f",
    "check": "#1ed75f",
    "normal": "#777777",
    "hint": "#00a6ff",
}


def listify_multiple(
    *values: str,
    key: str,
    pad: bool = True,
    join: bool = False,
    spacer: str = "",
    suffix: str = "",
) -> ty.List[str]:
    """Parse multiple inputs into single string."""
    ret = []
    for value in values:
        value = pad_str(value, suffix) if pad else f"{value!s}{suffix}"
        if join:
            ret.append(f"{key}{value}")
        else:
            ret.extend([key, value])
    return ret


def pad_str(value: ty.Any, suffix: str = "") -> str:
    """Pad string with quotes around out."""
    if IS_WIN:
        return f'"{value!s}{suffix}"'
    return f"{value!s}{suffix}"


def get_icon_state(errors: list[str]) -> tuple[str, str]:
    """Return icon state based on the error messages."""
    state = "check"
    if len([error for error in errors if 'class="warning"' in error]) > 0:
        state = "check_warning"
    if len([error for error in errors if 'class="error"' in error]) > 0:
        state = "error"
    return state, COLORS[state]


def format_interval(t: float) -> str:
    """
    Formats a number of seconds as a clock time, [H:]MM:SS.

    Parameters
    ----------
    t  : float
        Number of seconds.

    Returns
    -------
    out  : str
        [H:]MM:SS
    """
    if t is None:
        return "N/A"
    minutes, s = divmod(int(t), 60)
    h, m = divmod(minutes, 60)
    if h:
        return f"{h:d}:{m:02d}:{s:02d}"
    return f"{m:02d}:{s:02d}"


def format_timestamp(timestamp: float) -> str:
    """Format timestamp so it returns a human-readable string."""
    import datetime

    if not timestamp:
        return "N/A"
    return datetime.datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d %H:%M:%S")


def iterable_callbacks(func: ty.Optional[ty.Union[ty.Callable, Callback]]) -> ty.Sequence[ty.Callable]:
    """Callbacks should always be a sequence."""
    if func is None:
        return []
    elif isinstance(func, ty.Sequence):
        return func
    return [func]


def _safe_call(func: ty.Optional[ty.Sequence[ty.Callable]], task: ty.Optional["Task"] = None, which: str = "") -> None:
    if func is not None:
        try:
            for _func in func:
                with suppress(Exception):
                    _func() if task is None else _func(task)
        except Exception as err:
            with suppress(Exception):
                print(f"Exception raised in call (source={which}): {err}; func={func}; task={task}")


# 7-bit C1 ANSI sequences
ansi_escape = re.compile(
    r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
"""
)


def escape_ansi(text: str) -> str:
    """Try auto-escaping ANSI codes.

    See: https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
    """
    try:
        return ansi_escape.sub("", text)
    except Exception:
        return text


def format_command(commands: list[str], is_dev: bool = True) -> str:
    """Format command."""
    command = "; ".join(commands)
    if is_dev:
        command = command.replace("--no_color --debug", "--dev")
    else:
        command = command.replace("--no_color --debug", "--debug")
    return command



"""Info widget."""

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QFormLayout, QPlainTextEdit, QWidget

import qtextra.helpers as hp
from qtextra.widgets.qt_dialog import QtFramelessPopup


class InfoDialog(QtFramelessPopup):
    """Dialog to display some useful information."""

    def __init__(
        self,
        parent: QWidget,
        text: str,
        title: str = "Help information",
        min_width: int = 300,
        min_height: int = 200,
        persist: bool = False,
        quick: bool = False,
    ):
        self._title = title
        self.persist = persist
        self.quick = quick
        kws = {}
        if persist:
            kws = {"flags": Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowStaysOnTopHint | Qt.WindowType.Tool}
        super().__init__(parent, **kws)
        self.setMinimumWidth(min_width)
        self.setMinimumHeight(min_height)
        if self.quick:
            self.label.appendHtml(text)
            self.on_scroll_to_top()
        else:
            self.label.setText(text)

    def on_scroll_to_top(self) -> None:
        """Scroll to end."""
        self.label.verticalScrollBar().setValue(self.label.verticalScrollBar().minimum())

    def on_scroll_to_end(self) -> None:
        """Scroll to end."""
        self.label.verticalScrollBar().setValue(self.label.verticalScrollBar().maximum())

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QFormLayout:
        """Make panel."""
        if self.persist:
            _, header_layout = self._make_close_handle(self._title)
        else:
            header_layout = self._make_title_handle(self._title)

        layout = hp.make_form_layout(parent=self, margin=2)
        layout.addRow(header_layout)

        if self.quick:
            self.label = QPlainTextEdit()
            layout.addRow(
                hp.make_h_layout(
                    hp.make_btn(self, "Scroll to top", func=self.on_scroll_to_top),
                    hp.make_btn(self, "Scroll to end", func=self.on_scroll_to_end),
                    stretch_after=True,
                )
            )
        else:
            self.label = hp.make_scrollable_label(
                self, wrap=True, enable_url=True, selectable=True, object_name="title_label"
            )

        layout.addRow(self.label)
        return layout


if __name__ == "__main__":
    import sys

    from qtextra.utils.dev import apply_style, qapplication

    _ = qapplication()  # analysis:ignore

    dlg = InfoDialog(None, "This is a test message.<br>" * 50, persist=True, quick=True)
    apply_style(dlg)
    dlg.show()
    sys.exit(dlg.exec_())



"""Qt Logging module.

Copied and modified from this post: http://plumberjack.blogspot.com/2019/11/a-qt-gui-for-logging.html
"""

from __future__ import annotations

import typing as ty
from contextlib import suppress

from koyo.logging import LOG_FMT
from koyo.typing import PathLike
from loguru import logger
from qtpy.QtCore import QObject, Qt, Signal, Slot
from qtpy.QtGui import QFont
from qtpy.QtWidgets import QHBoxLayout, QPlainTextEdit, QVBoxLayout, QWidget

from qtextra.utils.utilities import connect
from qtextra.widgets.qt_dialog import QtFramelessTool
from qtextra.widgets.qt_toolbar_mini import QtMiniToolbar


class QtHandler(QObject):
    """Qt Log Handler."""

    evt_signal = Signal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def write(self, message: str):
        """Emit message."""
        with suppress(RuntimeError):
            self.evt_signal.emit(message)


class QtLogger(QWidget):
    """Logger window."""

    THEME_COLORS: ty.ClassVar[dict] = {
        "light": {
            "BACKGROUND": "white",
            "TEXT": "black",
            "TRACE": "darkblue",
            "DEBUG": "green",
            "INFO": "black",
            "WARNING": "orange",
            "ERROR": "red",
            "CRITICAL": "darkred",
        },
        "dark": {
            "BACKGROUND": "black",
            "TEXT": "white",
            "TRACE": "lightblue",
            "DEBUG": "green",
            "INFO": "white",
            "WARNING": "orange",
            "ERROR": "red",
            "CRITICAL": "darkred",
        },
    }
    THEMES: ty.ClassVar[list] = list(THEME_COLORS.keys())
    LOG_LEVELS: ty.ClassVar[dict] = {
        "[TRACE ": "TRACE",
        "[DEBUG ": "DEBUG",
        "[INFO ": "INFO",
        "[WARNING ": "WARNING",
        "[ERROR ": "ERROR",
        "[CRITICAL ": "CRITICAL",
    }
    THEME: str
    COLORS: ty.Dict[str, str]
    TEXT_COLOR: str

    def __init__(self, parent: QWidget | None = None, log_dir: PathLike | None = None):
        super().__init__(parent)
        self.log_dir = log_dir

        # Set whatever the default monospace font is for the platform
        self.textedit = QPlainTextEdit(self)
        font = QFont("monospace")
        self.textedit.setFont(font)
        self.textedit.setReadOnly(True)

        toolbar = QtMiniToolbar(self, Qt.Orientation.Vertical)
        toolbar.insert_qta_tool(
            "folder", tooltip="Open directory with logs", func=self.on_open_log_dir, hidden=self.log_dir is None
        )
        toolbar.insert_qta_tool("paint_palette", tooltip="Change theme", func=self.swap_theme)

        # Lay out all the widgets
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.textedit, stretch=1)
        layout.addWidget(toolbar)
        layout.setSpacing(0)

        # setup theme
        self.set_theme("dark")

        # setup logging
        self.handler = QtHandler(parent=self)
        connect(self.handler.evt_signal, self.update_log, state=True)
        self.log_id = logger.add(
            self.handler, level=0, backtrace=True, diagnose=True, catch=True, enqueue=True, format=LOG_FMT
        )

    def closeEvent(self, evt):
        """Close."""
        with suppress(Exception):
            logger.remove(self.log_id)
        connect(self.handler.evt_signal, self.update_log, state=False)
        return super().closeEvent(evt)

    @Slot(str)
    @Slot(object)
    def update_log(self, message):
        """Update log record."""
        record = message.record
        level = record["level"].name if hasattr(record["level"], "name") else record["level"]
        color = self.COLORS.get(level, self.TEXT_COLOR)
        self.append_log_entry(color, message)

    def append_log_entry(self, color: str | list, message: str) -> None:
        """Add log entry to the window."""
        self.textedit.appendHtml(f'<pre><font color="{color}">{message}</font></pre>')
        # self.textedit.appendHtml(f'<span><font color="{color}">{message}</font></span>')

    def on_open_log_dir(self):
        """Open directory containing log files."""
        from koyo.path import open_directory_alt

        if self.log_dir is None:
            logger.warning("Log directory is not set.")
            return
        open_directory_alt(self.log_dir)

    def swap_theme(self):
        """Swap theme."""
        current_theme = self.THEME
        new_theme = "light" if current_theme == "dark" else "dark"
        self.set_theme(new_theme)

    def set_theme(self, theme):
        """Update color theme of the logger."""
        if theme not in self.THEMES:
            raise ValueError(f"Theme must be one of `{self.THEMES}`")

        self.COLORS = self.THEME_COLORS[theme]
        self.TEXT_COLOR = self.THEME_COLORS[theme]["TEXT"]
        self.THEME = theme

        # update background color
        self.textedit.setStyleSheet("QPlainTextEdit {{background-color: {}}}".format(self.COLORS["BACKGROUND"]))
        self.recolor_old_text()

    def recolor_old_text(self):
        """Iteratively go through each line of the previous log and change the color to match the current theme."""

        def _get_log_level():
            for _level in self.LOG_LEVELS.keys():
                if _level in status:
                    return self.LOG_LEVELS[_level]

        old_log = self.textedit.toPlainText()
        old_log = old_log.split("\n")
        self.textedit.clear()

        for status in old_log:
            level = _get_log_level()
            color = self.COLORS.get(level, self.TEXT_COLOR)
            self.append_log_entry(color, status)


class QtLoggerDialog(QtFramelessTool):
    """Logger popup."""

    HIDE_WHEN_CLOSE = True

    def __init__(self, parent: ty.Optional[QWidget] = None, log_dir: PathLike | None = None) -> None:
        self.log_dir = log_dir
        super().__init__(parent)
        self.setMinimumWidth(600)
        self.setMinimumHeight(400)
        self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)  # type: ignore

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QVBoxLayout:
        """Make panel."""
        _, header_layout = self._make_hide_handle()
        self._title_label.setText("Logger")

        self.logger = QtLogger(self, log_dir=self.log_dir)

        layout = QVBoxLayout()
        layout.addLayout(header_layout)
        layout.addWidget(self.logger, stretch=True)
        layout.setSpacing(2)
        layout.setContentsMargins(6, 6, 6, 6)
        return layout


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import apply_style, qapplication

    app = qapplication(1)
    dlg = QtLoggerDialog()
    apply_style(dlg)
    dlg.show()

    logger.trace("TRACE")
    logger.debug("DEBUG")
    logger.info("INFO")
    logger.warning("WARNING")
    logger.error("ERROR")
    logger.critical("CRITICAL")

    sys.exit(app.exec_())



"""Sentry monitoring service."""

from __future__ import annotations

import typing as ty
from contextlib import suppress
from pathlib import Path

import sentry_sdk
from qtpy.QtWidgets import QWidget

from qtextra.dialogs.sentry.feedback import FeedbackDialog
from qtextra.dialogs.sentry.telemetry import TelemetryOptInDialog
from qtextra.dialogs.sentry.utilities import SENTRY_SETTINGS, _get_tags, get_sample_event

INSTALLED = False


__all__ = [
    "FeedbackDialog",
    "TelemetryOptInDialog",
    "ask_opt_in",
    "capture_exception",
    "get_sample_event",
    "install_error_monitor",
]

capture_exception = sentry_sdk.capture_exception


class Settings(ty.Protocol):
    """Settings protocol."""

    telemetry_enabled: bool
    telemetry_with_locals: bool


def ask_opt_in(settings: Settings, force: bool = False, parent: QWidget | None = None) -> Settings:
    """Show the dialog asking the user to opt in.

    Parameters
    ----------
    settings
        Settings object.
    force : bool, optional
        If True, will show opt_in even if user has already opted in/out,
        by default False.
    parent : QWidget, optional
        Parent widget, by default None.

    Returns
    -------
    SettingsDict
        [description].
    """
    assert settings is not None, "Settings must be provided."

    enabled_attr = "enabled" if hasattr(settings, "enabled") else "telemetry_enabled"
    with_locals_attr = "with_locals" if hasattr(settings, "with_locals") else "telemetry_with_locals"

    assert hasattr(settings, enabled_attr), f"Settings must have '{enabled_attr}' attribute."
    assert hasattr(settings, with_locals_attr), f"Settings must have '{with_locals_attr}' attribute."

    if not force and getattr(settings, enabled_attr) is not None:
        return settings

    dlg = TelemetryOptInDialog(parent=parent, with_locals=getattr(settings, with_locals_attr))
    send: ty.Optional[bool] = None
    if bool(dlg.exec()):
        send = True  # pragma: no cover
    elif dlg._no:
        send = False  # pragma: no cover

    if send is not None:
        setattr(settings, enabled_attr, send)
        setattr(settings, with_locals_attr, dlg._send_locals)
    return settings


def install_error_monitor(settings: Settings, **extra_kws: ty.Any) -> None:
    """Initialize the error monitor with sentry.io."""
    global INSTALLED
    if INSTALLED:
        return

    enabled_attr = "enabled" if hasattr(settings, "enabled") else "telemetry_enabled"
    with_locals_attr = "with_locals" if hasattr(settings, "with_locals") else "telemetry_with_locals"

    settings = ask_opt_in(settings)
    if not getattr(settings, enabled_attr):
        return

    _settings = SENTRY_SETTINGS.copy()
    _settings["include_local_variables"] = getattr(settings, with_locals_attr)
    sentry_sdk.init(**_settings)
    for k, v in _get_tags().items():
        sentry_sdk.set_tag(k, v)
    if extra_kws:
        set_extra_tags(**extra_kws)
    INSTALLED = True


def set_extra_tags(**kwargs: ty.Any) -> None:
    """Set extra tags."""
    for k, v in kwargs.items():
        sentry_sdk.set_tag(k, v)


def report_memory_usage() -> str:
    """Report memory usage for current process."""
    import tempfile

    import psutil
    from koyo.faulthandler import submit_sentry_attachment
    from koyo.utilities import human_readable_byte_size

    tmp = Path(tempfile.gettempdir())
    path = tmp / "memory_usage.txt"
    memory_text = f"Total memory: {human_readable_byte_size(psutil.virtual_memory().total)}\n"
    memory_text += f"Available memory: {human_readable_byte_size(psutil.virtual_memory().available)}\n"
    with suppress(Exception):
        process = psutil.Process()
        pid = process.pid
        mem = process.memory_info()

        memory_text += f"RMS ({pid}): {human_readable_byte_size(mem.rss)}\n"
        memory_text += f"VMS ({pid}): {human_readable_byte_size(mem.rss)}\n"
    path.write_text(memory_text)
    submit_sentry_attachment("Current memory usage - encountered memory issues", path)
    with suppress(Exception):
        path.unlink()
    return memory_text



"""Simple color scheme builder."""

from __future__ import annotations

import typing as ty
from copy import deepcopy

from koyo.color import colormap_to_hex
from pydantic.color import Color
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QDialogButtonBox, QHBoxLayout, QLayout, QWidget

import qtextra.helpers as hp
from qtextra.widgets.qt_dialog import QtDialog


def parse_colors(colors: ty.Union[ty.List[str], ty.List[Color]]) -> ty.List[str]:
    """Parse colors."""
    _colors = []
    for color in colors:
        if isinstance(color, Color):
            _colors.append(color.as_hex())
        else:
            _colors.append(color)
    return _colors


class QtColorListDialog(QtDialog):
    """Config import."""

    def __init__(
        self,
        parent: QWidget | None,
        colors: ty.Union[ty.List[str], ty.List[Color]],
        message: str = "Please click on any of the colors and select a new one",
    ):
        self.colors: ty.List[str] = parse_colors(colors)
        self.new_colors: ty.List[str] = deepcopy(self.colors)
        self.message = message
        super().__init__(parent)

    @property
    def n_colors(self) -> int:
        """Return the number of colors."""
        return len(self.colors)

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QLayout:
        """Make panel."""
        self.info_label = hp.make_label(self, self.message)

        color_layout, self.swatches = hp.make_swatch_grid(self, self.colors, self.on_update_color, use_flow_layout=True)

        colormap = hp.make_label(self, "Colormap")
        self.colormap = hp.make_combobox(
            self,
            ["custom", "viridis", "inferno", "magma", "plasma", "cividis", "twilight"],
        )
        self.colormap.currentTextChanged.connect(self.on_set_colormap)
        self.randomize_btn = hp.make_qta_btn(self, "shuffle", tooltip="Randomize colors", medium=False)

        self.randomize_btn.clicked.connect(self.on_randomize)
        self.invert = hp.make_checkbox(self, "Invert", tooltip="Reverse colors")
        self.invert.stateChanged.connect(self.on_set_colormap)
        hp.disable_widgets(self.invert, disabled=self.colormap.currentText() == "custom")

        layout = QHBoxLayout()
        layout.addWidget(self.colormap, stretch=1)
        layout.addWidget(self.randomize_btn)
        layout.addWidget(self.invert)

        # buttons
        self.button_box = QDialogButtonBox(self)
        self.button_box.setStandardButtons(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        self.button_box.setCenterButtons(True)

        # bind events
        self.button_box.button(QDialogButtonBox.StandardButton.Ok).clicked.connect(self.accept)
        self.button_box.button(QDialogButtonBox.StandardButton.Cancel).clicked.connect(self.reject)

        # set layout
        vertical_layout = hp.make_form_layout()
        vertical_layout.addRow(self.info_label)
        vertical_layout.addRow(colormap, layout)
        vertical_layout.addRow(color_layout)
        vertical_layout.addRow(self.button_box)

        return vertical_layout

    def on_randomize(self) -> None:
        """Randomize colors."""
        from koyo.color import get_random_hex_color

        self.colors = [get_random_hex_color() for _ in range(self.n_colors)]
        if self.colormap.currentText() == "custom":
            self.on_set_colormap()
        else:
            self.colormap.setCurrentText("custom")

    @Slot()  # type: ignore[misc]
    def on_set_colormap(self) -> None:
        """Set colors based on colormap."""
        import matplotlib.cm

        colormap = self.colormap.currentText()
        hp.disable_widgets(self.invert, disabled=colormap == "custom")
        if colormap == "custom":
            colors = self.colors
        else:
            colormap += "_r" if self.invert.isChecked() else ""
            cmap = matplotlib.colormaps.get_cmap(colormap)
            colors = colormap_to_hex(cmap.resampled(self.n_colors))
        for color_idx, (swatch, color) in enumerate(zip(self.swatches, colors)):
            swatch.set_color(color)
            self.new_colors[color_idx] = color

    def on_update_color(self, color_idx: int, color: str) -> None:
        """Update color."""
        self.new_colors[color_idx] = color

    def accept(self) -> None:
        """Accept changes."""
        self.colors = self.new_colors
        super().accept()


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import apply_style, qapplication

        _ = qapplication()  # analysis:ignore
        dlg = QtColorListDialog(
            None,
            [
                "#ff0000",
                "#00ff00",
                "#000075",
                "#a9a9a9",
                "#ff0000",
                "#00ff00",
                "#000075",
                "#a9a9a9",
                "#ff0000",
                "#00ff00",
                "#000075",
                "#a9a9a9",
            ],
        )
        apply_style(dlg)
        dlg.show()
        sys.exit(dlg.exec_())  # type: ignore[attr-defined]

    _main()  # type: ignore[no-untyped-call]
